import os
import sys
import time
from pathlib import Path
import pkgutil
import math
from math import erfc

import numpy as np
from astropy.table import Table
from astropy.io import ascii

import numba
from numba import njit, types, jit
from numba.typed import Dict

# import yaml
# config = yaml.safe_load(open('config/abacus_hod.yaml'))
# numba.set_num_threads(16)
float_array = types.float64[:]
int_array = types.int64[:]

@njit(fastmath=True)
def n_sat_LRG_modified(M_h, logM_cut, M_cut, M_1, sigma, alpha, kappa): 
    """
    Standard Zheng et al. (2005) satellite HOD parametrization for LRGs, modified with n_cent_LRG
    """
    if M_h - kappa*M_cut < 0:
        return 0
    return ((M_h - kappa*M_cut)/M_1)**alpha*0.5*math.erfc((logM_cut - np.log10(M_h))/(1.41421356*sigma))


@njit(fastmath=True)
def n_cen_LRG(M_h, logM_cut, sigma): 
    """
    Standard Zheng et al. (2005) central HOD parametrization for LRGs.
    """
    return 0.5*math.erfc((logM_cut - np.log10(M_h))/(1.41421356*sigma))

@njit(fastmath=True)
def N_sat_generic(M_h, M_cut, kappa, M_1, alpha, A_s=1.):
    """
    Standard Zheng et al. (2005) satellite HOD parametrization for all tracers with an optional amplitude parameter, A_s.
    """
    if M_h - kappa*M_cut < 0:
        return 0
    return A_s*((M_h-kappa*M_cut)/M_1)**alpha

@njit(fastmath=True)
def N_cen_ELG_v1(M_h, p_max, Q, logM_cut, sigma, gamma, Anorm = 1):
    """
    HOD function for ELG centrals taken from arXiv:1910.05095.
    """
    logM_h = np.log10(M_h)
    phi = phi_fun(logM_h, logM_cut, sigma)
    Phi = Phi_fun(logM_h, logM_cut, sigma, gamma)
    A = A_fun(p_max, Q, phi, Phi)
    return 2.*A*phi*Phi/Anorm + 0.5/Q*(1 + math.erf((logM_h-logM_cut)*100))

@njit(fastmath=True)
def N_cen_ELG_v2(M_h, p_max, logM_cut, sigma, gamma):
    """
    HOD function for ELG centrals taken from arXiv:2007.09012.
    """
    logM_h = np.log10(M_h)
    if logM_h <= logM_cut:
        return p_max*Gaussian_fun(logM_h, logM_cut, sigma)
    else:
        return p_max*(M_h/10**logM_cut)**gamma/(2.5066283*sigma)

@njit(fastmath=True)
def N_cen_QSO(M_h, p_max, logM_cut, sigma):
    """
    HOD function (Zheng et al. (2005) with p_max) for QSO centrals taken from arXiv:2007.09012.
    """
    return 0.5*p_max*(1 + math.erf((np.log10(M_h)-logM_cut)/1.41421356/sigma))


@njit(fastmath=True)
def phi_fun(logM_h, logM_cut, sigma):
    """
    Aiding function for N_cen_ELG_v1().
    """
    phi = Gaussian_fun(logM_h, logM_cut, sigma)
    return phi

@njit(fastmath=True)
def Phi_fun(logM_h, logM_cut, sigma, gamma):
    """
    Aiding function for N_cen_ELG_v1().
    """
    x = gamma*(logM_h-logM_cut)/sigma
    Phi = 0.5*(1 + math.erf(x/np.sqrt(2)))
    return Phi
    
@njit(fastmath=True)
def A_fun(p_max, Q, phi, Phi):
    """
    Aiding function for N_cen_ELG_v1().
    """
    A = (p_max-1./Q)
    return A
    
@njit(fastmath=True)
def Gaussian_fun(x, mean, sigma):
    """
    Gaussian function with centered at `mean' with standard deviation `sigma'.
    """
    return 0.3989422804014327/sigma*np.exp(-(x - mean)**2/2/sigma**2)


@njit(fastmath=True)
def wrap(x, L):
    '''Fast scalar mod implementation'''
    L2 = L/2
    if x >= L2:
        return x - L
    elif x < -L2:
        return x + L
    return x


@njit(parallel=True, fastmath=True)
def gen_cent(pos, vel, mass, ids, multis, randoms, vdev, deltac, fenv, 
    LRG_design_array, LRG_decorations_array, ELG_design_array, 
    ELG_decorations_array, QSO_design_array, QSO_decorations_array, 
    rsd, inv_velz2kms, lbox, want_LRG, want_ELG, want_QSO, Nthread):
    """
    Generate central galaxies in place in memory with a two pass numba parallel implementation. 
    """

    # parse out the hod parameters 
    logM_cut_L, logM1_L, sigma_L, alpha_L, kappa_L = \
        LRG_design_array[0], LRG_design_array[1], LRG_design_array[2], LRG_design_array[3], LRG_design_array[4]
    ic_L, alpha_c_L, Ac_L, Bc_L = LRG_decorations_array[10], LRG_decorations_array[0], \
        LRG_decorations_array[6], LRG_decorations_array[8]

    pmax_E, Q_E, logM_cut_E, kappa_E, sigma_E, logM1_E, alpha_E, gamma_E, A_E = \
        ELG_design_array[0], ELG_design_array[1], ELG_design_array[2], ELG_design_array[3], ELG_design_array[4],\
        ELG_design_array[5], ELG_design_array[6], ELG_design_array[7], ELG_design_array[8]
    alpha_c_E, Ac_E, Bc_E, ic_E = ELG_decorations_array[0], ELG_decorations_array[6], ELG_decorations_array[8],\
    ELG_decorations_array[10]

    pmax_Q, logM_cut_Q, kappa_Q, sigma_Q, logM1_Q, alpha_Q, A_Q = \
        QSO_design_array[0], QSO_design_array[1], QSO_design_array[2], QSO_design_array[3], QSO_design_array[4],\
        QSO_design_array[5], QSO_design_array[6]
    alpha_c_Q, Ac_Q, Bc_Q, ic_Q = QSO_decorations_array[0], QSO_decorations_array[6], QSO_decorations_array[8],\
    QSO_decorations_array[10]

    H = len(mass)

    numba.set_num_threads(Nthread)
    Nout = np.zeros((Nthread, 3, 8), dtype = np.int64)
    hstart = np.rint(np.linspace(0, H, Nthread + 1)) # starting index of each thread

    keep = np.empty(H, dtype = np.int8) # mask array tracking which halos to keep

    # figuring out the number of halos kept for each thread
    for tid in numba.prange(Nthread):
        for i in range(hstart[tid], hstart[tid + 1]):
            # first create the markers between 0 and 1 for different tracers
            LRG_marker = 0
            if want_LRG:
                # do assembly bias and secondary bias
                logM_cut_L_temp = logM_cut_L + Ac_L * deltac[i] + Bc_L * fenv[i]
                LRG_marker += n_cen_LRG(mass[i], logM_cut_L_temp, sigma_L) * ic_L * multis[i]
            ELG_marker = LRG_marker
            if want_ELG:
                logM_cut_E_temp = logM_cut_E + Ac_E * deltac[i] + Bc_E * fenv[i]
                ELG_marker += N_cen_ELG_v1(mass[i], pmax_E, Q_E, logM_cut_E_temp, sigma_E, gamma_E, A_E) * ic_E * multis[i]
            QSO_marker = ELG_marker
            if want_QSO:
                logM_cut_Q_temp = logM_cut_Q + Ac_Q * deltac[i] + Bc_Q * fenv[i]
                QSO_marker += N_cen_QSO(mass[i], pmax_Q, logM_cut_Q, sigma_Q) * ic_Q * multis[i]

            if randoms[i] <= LRG_marker:
                Nout[tid, 0, 0] += 1 # counting
                keep[i] = 1
            elif randoms[i] <= ELG_marker:
                Nout[tid, 1, 0] += 1 # counting
                keep[i] = 2
            elif randoms[i] <= QSO_marker:
                Nout[tid, 2, 0] += 1 # counting
                keep[i] = 3
            else:
                keep[i] = 0

    # compose galaxy array, first create array of galaxy starting indices for the threads
    gstart = np.empty((Nthread + 1, 3), dtype = np.int64)
    gstart[0, :] = 0
    gstart[1:, 0] = Nout[:, 0, 0].cumsum()
    gstart[1:, 1] = Nout[:, 1, 0].cumsum()
    gstart[1:, 2] = Nout[:, 2, 0].cumsum()

    # galaxy arrays
    N_lrg = gstart[-1, 0]
    lrg_x = np.empty(N_lrg, dtype = mass.dtype)
    lrg_y = np.empty(N_lrg, dtype = mass.dtype)
    lrg_z = np.empty(N_lrg, dtype = mass.dtype)
    lrg_vx = np.empty(N_lrg, dtype = mass.dtype)
    lrg_vy = np.empty(N_lrg, dtype = mass.dtype)
    lrg_vz = np.empty(N_lrg, dtype = mass.dtype)
    lrg_mass = np.empty(N_lrg, dtype = mass.dtype)
    lrg_id = np.empty(N_lrg, dtype = ids.dtype)

    # galaxy arrays
    N_elg = gstart[-1, 1]
    elg_x = np.empty(N_elg, dtype = mass.dtype)
    elg_y = np.empty(N_elg, dtype = mass.dtype)
    elg_z = np.empty(N_elg, dtype = mass.dtype)
    elg_vx = np.empty(N_elg, dtype = mass.dtype)
    elg_vy = np.empty(N_elg, dtype = mass.dtype)
    elg_vz = np.empty(N_elg, dtype = mass.dtype)
    elg_mass = np.empty(N_elg, dtype = mass.dtype)
    elg_id = np.empty(N_elg, dtype = ids.dtype)

    # galaxy arrays
    N_qso = gstart[-1, 2]
    qso_x = np.empty(N_qso, dtype = mass.dtype)
    qso_y = np.empty(N_qso, dtype = mass.dtype)
    qso_z = np.empty(N_qso, dtype = mass.dtype)
    qso_vx = np.empty(N_qso, dtype = mass.dtype)
    qso_vy = np.empty(N_qso, dtype = mass.dtype)
    qso_vz = np.empty(N_qso, dtype = mass.dtype)
    qso_mass = np.empty(N_qso, dtype = mass.dtype)
    qso_id = np.empty(N_qso, dtype = ids.dtype)

    # fill in the galaxy arrays
    for tid in numba.prange(Nthread):
        j1, j2, j3 = gstart[tid]
        for i in range(hstart[tid], hstart[tid + 1]):
            if keep[i] == 1:
                # loop thru three directions to assign galaxy velocities and positions
                lrg_x[j1] = pos[i,0]
                lrg_vx[j1] = vel[i,0] + alpha_c_L * vdev[i] # velocity bias
                lrg_y[j1] = pos[i,1]
                lrg_vy[j1] = vel[i,1] + alpha_c_L * vdev[i] # velocity bias
                lrg_z[j1] = pos[i,2]
                lrg_vz[j1] = vel[i,2] + alpha_c_L * vdev[i] # velocity bias
                # rsd only applies to the z direction
                if rsd:
                    lrg_z[j1] = wrap(pos[i,2] + lrg_vz[j1] * inv_velz2kms, lbox)
                lrg_mass[j1] = mass[i]
                lrg_id[j1] = ids[i]
                j1 += 1
            elif keep[i] == 2:
                # loop thru three directions to assign galaxy velocities and positions
                elg_x[j2] = pos[i,0]
                elg_vx[j2] = vel[i,0] + alpha_c_E * vdev[i] # velocity bias
                elg_y[j2] = pos[i,1]
                elg_vy[j2] = vel[i,1] + alpha_c_E * vdev[i] # velocity bias
                elg_z[j2] = pos[i,2]
                elg_vz[j2] = vel[i,2] + alpha_c_E * vdev[i] # velocity bias
                # rsd only applies to the z direction
                if rsd:
                    elg_z[j2] = wrap(pos[i,2] + elg_vz[j2] * inv_velz2kms, lbox)
                elg_mass[j2] = mass[i]
                elg_id[j2] = ids[i]
                j2 += 1
            elif keep[i] == 3:
                # loop thru three directions to assign galaxy velocities and positions
                qso_x[j3] = pos[i,0]
                qso_vx[j3] = vel[i,0] + alpha_c_Q * vdev[i] # velocity bias
                qso_y[j3] = pos[i,1]
                qso_vy[j3] = vel[i,1] + alpha_c_Q * vdev[i] # velocity bias
                qso_z[j3] = pos[i,2]
                qso_vz[j3] = vel[i,2] + alpha_c_Q * vdev[i] # velocity bias
                # rsd only applies to the z direction
                if rsd:
                    qso_z[j3] = wrap(pos[i,2] + qso_vz[j3] * inv_velz2kms, lbox)
                qso_mass[j3] = mass[i]
                qso_id[j3] = ids[i]
                j3 += 1
        # assert j == gstart[tid + 1]

    LRG_dict = Dict.empty(key_type = types.unicode_type, value_type = float_array)
    ELG_dict = Dict.empty(key_type = types.unicode_type, value_type = float_array)
    QSO_dict = Dict.empty(key_type = types.unicode_type, value_type = float_array)
    ID_dict = Dict.empty(key_type = types.unicode_type, value_type = int_array)
    LRG_dict['x'] = lrg_x
    LRG_dict['y'] = lrg_y
    LRG_dict['z'] = lrg_z
    LRG_dict['vx'] = lrg_vx
    LRG_dict['vy'] = lrg_vy
    LRG_dict['vz'] = lrg_vz
    LRG_dict['mass'] = lrg_mass
    ID_dict['LRG'] = lrg_id

    ELG_dict['x'] = elg_x
    ELG_dict['y'] = elg_y
    ELG_dict['z'] = elg_z
    ELG_dict['vx'] = elg_vx
    ELG_dict['vy'] = elg_vy
    ELG_dict['vz'] = elg_vz
    ELG_dict['mass'] = elg_mass
    ID_dict['ELG'] = elg_id

    QSO_dict['x'] = qso_x
    QSO_dict['y'] = qso_y
    QSO_dict['z'] = qso_z
    QSO_dict['vx'] = qso_vx
    QSO_dict['vy'] = qso_vy
    QSO_dict['vz'] = qso_vz
    QSO_dict['mass'] = qso_mass
    ID_dict['QSO'] = qso_id
    return LRG_dict, ELG_dict, QSO_dict, ID_dict


@njit(parallel = True, fastmath = True)
def gen_sats(ppos, pvel, hvel, hmass, hid, weights, randoms, hdeltac, hfenv, 
    enable_ranks, ranks, ranksv, ranksp, ranksr, 
    LRG_design_array, LRG_decorations_array, ELG_design_array, ELG_decorations_array,
    QSO_design_array, QSO_decorations_array,
    rsd, inv_velz2kms, lbox, Mpart, want_LRG, want_ELG, want_QSO, Nthread):

    """
    Generate satellite galaxies in place in memory with a two pass numba parallel implementation. 
    """

    # standard hod design
    logM_cut_L, logM1_L, sigma_L, alpha_L, kappa_L = \
        LRG_design_array[0], LRG_design_array[1], LRG_design_array[2], LRG_design_array[3], LRG_design_array[4]
    alpha_s_L, s_L, s_v_L, s_p_L, s_r_L, Ac_L, As_L, Bc_L, Bs_L, ic_L = \
        LRG_decorations_array[1], LRG_decorations_array[2], LRG_decorations_array[3], LRG_decorations_array[4], \
        LRG_decorations_array[5], LRG_decorations_array[6], LRG_decorations_array[7], LRG_decorations_array[8], \
        LRG_decorations_array[9], LRG_decorations_array[10]

    pmax_E, Q_E, logM_cut_E, kappa_E, sigma_E, logM1_E, alpha_E, gamma_E, A_E = \
        ELG_design_array[0], ELG_design_array[1], ELG_design_array[2], ELG_design_array[3], ELG_design_array[4],\
        ELG_design_array[5], ELG_design_array[6], ELG_design_array[7], ELG_design_array[8]
    alpha_s_E, s_E, s_v_E, s_p_E, s_r_E, Ac_E, As_E, Bc_E, Bs_E, ic_E = \
        ELG_decorations_array[1], ELG_decorations_array[2], ELG_decorations_array[3], ELG_decorations_array[4], \
        ELG_decorations_array[5], ELG_decorations_array[6], ELG_decorations_array[7], ELG_decorations_array[8], \
        ELG_decorations_array[9], ELG_decorations_array[10]

    pmax_Q, logM_cut_Q, kappa_Q, sigma_Q, logM1_Q, alpha_Q, A_Q = \
        QSO_design_array[0], QSO_design_array[1], QSO_design_array[2], QSO_design_array[3], QSO_design_array[4],\
        QSO_design_array[5], QSO_design_array[6]
    alpha_s_Q, s_Q, s_v_Q, s_p_Q, s_r_Q, Ac_Q, As_Q, Bc_Q, Bs_Q, ic_Q = \
        QSO_decorations_array[1], QSO_decorations_array[2], QSO_decorations_array[3], QSO_decorations_array[4], \
        QSO_decorations_array[5], QSO_decorations_array[6], QSO_decorations_array[7], QSO_decorations_array[8], \
        QSO_decorations_array[9], QSO_decorations_array[10]

    H = len(hmass) # num of particles

    numba.set_num_threads(Nthread)
    Nout = np.zeros((Nthread, 3, 8), dtype = np.int64)
    hstart = np.rint(np.linspace(0, H, Nthread + 1)) # starting index of each thread

    keep = np.empty(H, dtype = np.int8) # mask array tracking which halos to keep

    # figuring out the number of particles kept for each thread
    for tid in numba.prange(Nthread): #numba.prange(Nthread):
        for i in range(hstart[tid], hstart[tid + 1]):
            # print(logM1, As, hdeltac[i], Bs, hfenv[i])
            LRG_marker = 0
            if want_LRG:
                M1_L_temp = 10**(logM1_L + As_L * hdeltac[i] + Bs_L * hfenv[i])
                logM_cut_L_temp = logM_cut_L + Ac_L * hdeltac[i] + Bc_L * hfenv[i]
                base_p_L = n_sat_LRG_modified(hmass[i], logM_cut_L_temp, 
                    10**logM_cut_L_temp, M1_L_temp, sigma_L, alpha_L, kappa_L) * weights[i] * ic_L
                if enable_ranks:
                    decorator_L = 1 + s_L * ranks[i] + s_v_L * ranksv[i] + s_p_L * ranksp[i] + s_r_L * ranksr[i]
                    exp_sat = base_p_L * decorator_L
                else:
                    exp_sat = base_p_L
                LRG_marker += exp_sat
            ELG_marker = LRG_marker
            if want_ELG:
                M1_E_temp = 10**(logM1_E + As_E * hdeltac[i] + Bs_E * hfenv[i])
                logM_cut_E_temp = logM_cut_E + Ac_E * hdeltac[i] + Bc_E * hfenv[i]
                base_p_E = N_sat_generic(
                    hmass[i], 10**logM_cut_E_temp, kappa_E, M1_E_temp, alpha_E) * weights[i] * ic_E
                if enable_ranks:
                    decorator_E = 1 + s_E * ranks[i] + s_v_E * ranksv[i] + s_p_E * ranksp[i] + s_r_E * ranksr[i]
                    exp_sat = base_p_E * decorator_E
                else:
                    exp_sat = base_p_E
                ELG_marker += exp_sat
            QSO_marker = ELG_marker
            if want_QSO:
                M1_Q_temp = 10**(logM1_Q + As_Q * hdeltac[i] + Bs_Q * hfenv[i])
                logM_cut_Q_temp = logM_cut_Q + Ac_Q * hdeltac[i] + Bc_Q * hfenv[i]
                base_p_Q = N_sat_generic(
                    hmass[i], 10**logM_cut_Q_temp, kappa_Q, M1_Q_temp, alpha_Q, A_Q) * weights[i] * ic_Q
                if enable_ranks:
                    decorator_Q = 1 + s_Q * ranks[i] + s_v_Q * ranksv[i] + s_p_Q * ranksp[i] + s_r_Q * ranksr[i]
                    exp_sat = base_p_Q * decorator_Q
                else:
                    exp_sat = base_p_Q
                QSO_marker += exp_sat

            if randoms[i] <= LRG_marker:
                Nout[tid, 0, 0] += 1 # counting
                keep[i] = 1
            elif randoms[i] <= ELG_marker:
                Nout[tid, 1, 0] += 1 # counting
                keep[i] = 2
            elif randoms[i] <= QSO_marker:
                Nout[tid, 2, 0] += 1 # counting
                keep[i] = 3    
            else:
                keep[i] = 0

    # compose galaxy array, first create array of galaxy starting indices for the threads
    gstart = np.empty((Nthread + 1, 3), dtype = np.int64)
    gstart[0, :] = 0
    gstart[1:, 0] = Nout[:, 0, 0].cumsum()
    gstart[1:, 1] = Nout[:, 1, 0].cumsum()
    gstart[1:, 2] = Nout[:, 2, 0].cumsum()

    # galaxy arrays
    N_lrg = gstart[-1, 0]
    lrg_x = np.empty(N_lrg, dtype = hmass.dtype)
    lrg_y = np.empty(N_lrg, dtype = hmass.dtype)
    lrg_z = np.empty(N_lrg, dtype = hmass.dtype)
    lrg_vx = np.empty(N_lrg, dtype = hmass.dtype)
    lrg_vy = np.empty(N_lrg, dtype = hmass.dtype)
    lrg_vz = np.empty(N_lrg, dtype = hmass.dtype)
    lrg_mass = np.empty(N_lrg, dtype = hmass.dtype)
    lrg_id = np.empty(N_lrg, dtype = hid.dtype)

    # galaxy arrays
    N_elg = gstart[-1, 1]
    elg_x = np.empty(N_elg, dtype = hmass.dtype)
    elg_y = np.empty(N_elg, dtype = hmass.dtype)
    elg_z = np.empty(N_elg, dtype = hmass.dtype)
    elg_vx = np.empty(N_elg, dtype = hmass.dtype)
    elg_vy = np.empty(N_elg, dtype = hmass.dtype)
    elg_vz = np.empty(N_elg, dtype = hmass.dtype)
    elg_mass = np.empty(N_elg, dtype = hmass.dtype)
    elg_id = np.empty(N_elg, dtype = hid.dtype)

    # galaxy arrays
    N_qso = gstart[-1, 2]
    qso_x = np.empty(N_qso, dtype = hmass.dtype)
    qso_y = np.empty(N_qso, dtype = hmass.dtype)
    qso_z = np.empty(N_qso, dtype = hmass.dtype)
    qso_vx = np.empty(N_qso, dtype = hmass.dtype)
    qso_vy = np.empty(N_qso, dtype = hmass.dtype)
    qso_vz = np.empty(N_qso, dtype = hmass.dtype)
    qso_mass = np.empty(N_qso, dtype = hmass.dtype)
    qso_id = np.empty(N_qso, dtype = hid.dtype)

    # fill in the galaxy arrays
    for tid in numba.prange(Nthread):
        j1, j2, j3 = gstart[tid]
        for i in range(hstart[tid], hstart[tid + 1]):
            if keep[i] == 1:
                lrg_x[j1] = ppos[i, 0]
                lrg_vx[j1] = hvel[i, 0] + alpha_s_L * (pvel[i, 0] - hvel[i, 0]) # velocity bias
                lrg_y[j1] = ppos[i, 1]
                lrg_vy[j1] = hvel[i, 1] + alpha_s_L * (pvel[i, 1] - hvel[i, 1]) # velocity bias
                lrg_z[j1] = ppos[i, 2]
                lrg_vz[j1] = hvel[i, 2] + alpha_s_L * (pvel[i, 2] - hvel[i, 2]) # velocity bias
                if rsd:
                    lrg_z[j1] = wrap(lrg_z[j1] + lrg_vz[j1] * inv_velz2kms, lbox)
                lrg_mass[j1] = hmass[i]
                lrg_id[j1] = hid[i]
                j1 += 1
            elif keep[i] == 2:
                elg_x[j2] = ppos[i, 0]
                elg_vx[j2] = hvel[i, 0] + alpha_s_E * (pvel[i, 0] - hvel[i, 0]) # velocity bias
                elg_y[j2] = ppos[i, 1]
                elg_vy[j2] = hvel[i, 1] + alpha_s_E * (pvel[i, 1] - hvel[i, 1]) # velocity bias
                elg_z[j2] = ppos[i, 2]
                elg_vz[j2] = hvel[i, 2] + alpha_s_E * (pvel[i, 2] - hvel[i, 2]) # velocity bias
                if rsd:
                    elg_z[j2] = wrap(elg_z[j2] + elg_vz[j2] * inv_velz2kms, lbox)
                elg_mass[j2] = hmass[i]
                elg_id[j2] = hid[i]
                j2 += 1
            elif keep[i] == 3:
                qso_x[j3] = ppos[i, 0]
                qso_vx[j3] = hvel[i, 0] + alpha_s_Q * (pvel[i, 0] - hvel[i, 0]) # velocity bias
                qso_y[j3] = ppos[i, 1]
                qso_vy[j3] = hvel[i, 1] + alpha_s_Q * (pvel[i, 1] - hvel[i, 1]) # velocity bias
                qso_z[j3] = ppos[i, 2]
                qso_vz[j3] = hvel[i, 2] + alpha_s_Q * (pvel[i, 2] - hvel[i, 2]) # velocity bias
                if rsd:
                    qso_z[j3] = wrap(qso_z[j3] + qso_vz[j3] * inv_velz2kms, lbox)
                qso_mass[j3] = hmass[i]
                qso_id[j3] = hid[i]
                j3 += 1
        # assert j == gstart[tid + 1]

    LRG_dict = Dict.empty(key_type = types.unicode_type, value_type = float_array)
    ELG_dict = Dict.empty(key_type = types.unicode_type, value_type = float_array)
    QSO_dict = Dict.empty(key_type = types.unicode_type, value_type = float_array)
    ID_dict = Dict.empty(key_type = types.unicode_type, value_type = int_array)
    LRG_dict['x'] = lrg_x
    LRG_dict['y'] = lrg_y
    LRG_dict['z'] = lrg_z
    LRG_dict['vx'] = lrg_vx
    LRG_dict['vy'] = lrg_vy
    LRG_dict['vz'] = lrg_vz
    LRG_dict['mass'] = lrg_mass
    ID_dict['LRG'] = lrg_id

    ELG_dict['x'] = elg_x
    ELG_dict['y'] = elg_y
    ELG_dict['z'] = elg_z
    ELG_dict['vx'] = elg_vx
    ELG_dict['vy'] = elg_vy
    ELG_dict['vz'] = elg_vz
    ELG_dict['mass'] = elg_mass
    ID_dict['ELG'] = elg_id

    QSO_dict['x'] = qso_x
    QSO_dict['y'] = qso_y
    QSO_dict['z'] = qso_z
    QSO_dict['vx'] = qso_vx
    QSO_dict['vy'] = qso_vy
    QSO_dict['vz'] = qso_vz
    QSO_dict['mass'] = qso_mass
    ID_dict['QSO'] = qso_id
    return LRG_dict, ELG_dict, QSO_dict, ID_dict

@njit(parallel = True, fastmath = True)
def fast_concatenate(array1, array2, Nthread):
    ''' Fast concatenate with numba parallel'''

    N1 = len(array1)
    N2 = len(array2)
    if N1 == 0:
        return array2
    elif N2 == 0:
        return array1

    final_array = np.empty(N1 + N2, dtype = array1.dtype)
    # if one thread, then no need to parallel
    if Nthread == 1:
        for i in range(N1):
            final_array[i] = array1[i]
        for j in range(N2):
            final_array[j + N1] = array2[j]
        return final_array

    numba.set_num_threads(Nthread)
    Nthread1 = max(1, int(np.floor(Nthread * N1 / (N1 + N2))))
    Nthread2 = Nthread - Nthread1
    hstart1 = np.rint(np.linspace(0, N1, Nthread1 + 1))
    hstart2 = np.rint(np.linspace(0, N2, Nthread2 + 1)) + N1

    for tid in numba.prange(Nthread): #numba.prange(Nthread):
        if tid < Nthread1:
            for i in range(hstart1[tid], hstart1[tid + 1]):
                final_array[i] = array1[i]
        else:
            for i in range(hstart2[tid - Nthread1], hstart2[tid + 1 - Nthread1]):
                final_array[i] = array2[i - N1]
    # final_array = np.concatenate((array1, array2))
    return final_array

def gen_gals(halos_array, subsample, tracers, params, Nthread, enable_ranks, rsd, verbose):
    """
    parse hod parameters, pass them on to central and satellite generators 
    and then format the results 

    Parameters
    ----------

    halos_array : dictionary of arrays 
        a dictionary of halo properties (pos, vel, mass, id, randoms, ...)

    subsample : dictionary of arrays
        a dictionary of particle propoerties (pos, vel, hmass, hid, Np, subsampling, randoms, ...)

    tracers : dictionary of dictionaries
        Dictionary of multi-tracer HODs

    enable_ranks : boolean
        Flag of whether to implement particle ranks. 

    rsd : boolean
        Flag of whether to implement RSD. 

    params : dict
        Dictionary of various simulation parameters. 

    """

    # B.H. TODO: pass as dictionary; make what's below more succinct
    for tracer in tracers.keys():
        if tracer == 'LRG':
            LRG_HOD = tracers[tracer]
        if tracer == 'ELG':
            ELG_HOD = tracers[tracer]
        if tracer == 'QSO':
            QSO_HOD = tracers[tracer]

    if 'LRG' in tracers.keys():
        want_LRG = True
        # LRG design and decorations
        logM_cut_L, logM1_L, sigma_L, alpha_L, kappa_L = \
            map(LRG_HOD.get, ('logM_cut', 
                              'logM1', 
                              'sigma', 
                              'alpha', 
                              'kappa'))
        LRG_design_array = np.array([logM_cut_L, logM1_L, sigma_L, alpha_L, kappa_L])

        alpha_c, alpha_s, s, s_v, s_p, s_r, Ac, As, Bc, Bs, ic = \
            map(LRG_HOD.get, ('alpha_c', 
                            'alpha_s',  
                            's', 
                            's_v', 
                            's_p', 
                            's_r',
                            'Acent',
                            'Asat',
                            'Bcent',
                            'Bsat',
                            'ic'))
        LRG_decorations_array = np.array([alpha_c, alpha_s, s, s_v, s_p, s_r, Ac, As, Bc, Bs, ic])
    else:
        # B.H. TODO: this will go when we switch to dictionaried and for loops
        want_LRG = False
        LRG_design_array = np.zeros(5)
        LRG_decorations_array = np.zeros(11)
        
    if 'ELG' in tracers.keys():
        # ELG design
        want_ELG = True
        pmax_E, Q_E, logM_cut_E, kappa_E, sigma_E, logM1_E, alpha_E, gamma_E, A_E = \
            map(ELG_HOD.get, ('p_max',
                            'Q',
                            'logM_cut',
                            'kappa',
                            'sigma',
                            'logM1',
                            'alpha',
                            'gamma',
                            'A_s'))
        ELG_design_array = np.array(
            [pmax_E, Q_E, logM_cut_E, kappa_E, sigma_E, logM1_E, alpha_E, gamma_E, A_E])
        alpha_c_E, alpha_s_E, s_E, s_v_E, s_p_E, s_r_E, Ac_E, As_E, Bc_E, Bs_E, ic_E = \
            map(ELG_HOD.get, ('alpha_c', 
                            'alpha_s',  
                            's', 
                            's_v', 
                            's_p', 
                            's_r',
                            'Acent',
                            'Asat',
                            'Bcent',
                            'Bsat',
                            'ic'))
        ELG_decorations_array = np.array(
            [alpha_c_E, alpha_s_E, s_E, s_v_E, s_p_E, s_r_E, Ac_E, As_E, Bc_E, Bs_E, ic_E])
    else:
        # B.H. TODO: this will go when we switch to dictionaried and for loops
        ELG_design_array = np.zeros(9)
        ELG_decorations_array = np.zeros(10)
        want_ELG = False
        
    if 'QSO' in tracers.keys():
        # QSO design
        want_QSO = True
        pmax_Q, logM_cut_Q, kappa_Q, sigma_Q, logM1_Q, alpha_Q, A_Q = \
            map(QSO_HOD.get, ('p_max',
                            'logM_cut',
                            'kappa',
                            'sigma',
                            'logM1',
                            'alpha',
                            'A_s'))
        QSO_design_array = np.array(
            [pmax_Q, logM_cut_Q, kappa_Q, sigma_Q, logM1_Q, alpha_Q, A_Q])
        alpha_c_Q, alpha_s_Q, s_Q, s_v_Q, s_p_Q, s_r_Q, Ac_Q, As_Q, Bc_Q, Bs_Q, ic_Q = \
            map(QSO_HOD.get, ('alpha_c', 
                            'alpha_s',  
                            's', 
                            's_v', 
                            's_p', 
                            's_r',
                            'Acent',
                            'Asat',
                            'Bcent',
                            'Bsat',
                            'ic'))
        QSO_decorations_array = np.array(
            [alpha_c_Q, alpha_s_Q, s_Q, s_v_Q, s_p_Q, s_r_Q, Ac_Q, As_Q, Bc_Q, Bs_Q, ic_Q])
    else:
        # B.H. TODO: this will go when we switch to dictionaried and for loops
        QSO_design_array = np.zeros(7)
        QSO_decorations_array = np.zeros(10)
        want_QSO = False
    
    start = time.time()

    velz2kms = params['velz2kms']
    inv_velz2kms = 1/velz2kms
    lbox = params['Lbox']
    # for each halo, generate central galaxies and output to file
    LRG_dict_cent, ELG_dict_cent, QSO_dict_cent, ID_dict_cent = \
    gen_cent(halos_array['hpos'], halos_array['hvel'], halos_array['hmass'], halos_array['hid'], halos_array['hmultis'], 
             halos_array['hrandoms'], halos_array['hveldev'], halos_array['hdeltac'], halos_array['hfenv'], 
             LRG_design_array, LRG_decorations_array, ELG_design_array, ELG_decorations_array, QSO_design_array, 
             QSO_decorations_array, rsd, inv_velz2kms, lbox, want_LRG, want_ELG, want_QSO, Nthread)
    if verbose:
        print("generating centrals took ", time.time() - start)


    start = time.time()
    LRG_dict_sat, ELG_dict_sat, QSO_dict_sat, ID_dict_sat = \
    gen_sats(subsample['ppos'], subsample['pvel'], subsample['phvel'], subsample['phmass'], subsample['phid'], 
             subsample['pweights'], subsample['prandoms'], subsample['pdeltac'], subsample['pfenv'], 
             enable_ranks, subsample['pranks'], subsample['pranksv'], subsample['pranksp'], subsample['pranksr'],
             LRG_design_array, LRG_decorations_array, ELG_design_array, ELG_decorations_array,
             QSO_design_array, QSO_decorations_array, rsd, inv_velz2kms, lbox, params['Mpart'],
             want_LRG, want_ELG, want_QSO, Nthread)
    if verbose:
        print("generating satellites took ", time.time() - start)

    # B.H. TODO: need a for loop above so we don't need to do this by hand
    HOD_dict_sat = {'LRG': LRG_dict_sat, 'ELG': ELG_dict_sat, 'QSO': QSO_dict_sat}
    HOD_dict_cent = {'LRG': LRG_dict_cent, 'ELG': ELG_dict_cent, 'QSO': QSO_dict_cent}
    
    # do a concatenate in numba parallel 
    start = time.time()
    HOD_dict = {}
    for tracer in tracers:
        tracer_dict = {'Ncent':len(HOD_dict_cent[tracer]['x'])}
        for k in HOD_dict_cent[tracer]:
            tracer_dict[k] = fast_concatenate(HOD_dict_cent[tracer][k], HOD_dict_sat[tracer][k], Nthread)
        tracer_dict['id'] = fast_concatenate(ID_dict_cent[tracer], ID_dict_sat[tracer], Nthread)
        if verbose:
            print(tracer, "number of galaxies ", len(tracer_dict['x']))
            print("satellite fraction ", len(HOD_dict_sat[tracer]['x'])/len(tracer_dict['x']))
        HOD_dict[tracer] = tracer_dict
    if verbose:
        print("organizing outputs took ", time.time() - start)
    return HOD_dict


def gen_gal_cat(halo_data, particle_data, tracers, params, Nthread = 16, 
    enable_ranks = False, rsd = True, write_to_disk = False, savedir = "./", verbose = False):
    """
    pass on inputs to the gen_gals function and takes care of I/O

    Parameters
    ----------

    halos_data : dictionary of arrays 
        a dictionary of halo properties (pos, vel, mass, id, randoms, ...)

    particle_data : dictionary of arrays
        a dictionary of particle propoerties (pos, vel, hmass, hid, Np, subsampling, randoms, ...)

    tracers : dictionary of dictionaries
        Dictionary of multi-tracer HODs

    enable_ranks : boolean
        Flag of whether to implement particle ranks. 

    rsd : boolean
        Flag of whether to implement RSD. 

    write_to_disk : boolean
        Flag of whether to output to disk. 

    verbose : boolean
        Whether to output detailed outputs. 

    savedir : str
        where to save the output if write_to_disk == True. 

    params : dict
        Dictionary of various simulation parameters. 

    Output
    ------

    HOD_dict : dictionary of dictionaries
        Dictionary of the format: {tracer1_dict, tracer2_dict, ...}, 
        where tracer1_dict = {x, y, z, vx, vy, vz, mass, id}

    """

    if not type(rsd) is bool:
        raise ValueError("Error: rsd has to be a boolean")

    # find the halos, populate them with galaxies and write them to files
    HOD_dict = gen_gals(halo_data, particle_data, tracers, params, Nthread, enable_ranks, rsd, verbose)
    
    # how many galaxies were generated and write them to disk
    for tracer in tracers.keys():
        Ncent = HOD_dict[tracer]['Ncent']
        if verbose:
            print("generated %ss:"%tracer, len(HOD_dict[tracer]['x']), 
                "satellite fraction ", 1 - Ncent/len(HOD_dict[tracer]['x']))
        
        if write_to_disk:
            if verbose:
                print("outputting galaxies to disk")

            if rsd:
                rsd_string = "_rsd"
            else:
                rsd_string = ""

            outdir = (savedir) / ("galaxies"+rsd_string)

            # create directories if not existing
            os.makedirs(outdir, exist_ok = True)

            # save to file 
            outdict = HOD_dict[tracer].pop('Ncent', None)
            table = Table(HOD_dict[tracer], meta = {'Ncent': Ncent, 'Gal_type': tracer, **tracers[tracer]})
            ascii.write(table, outdir / ("%ss.dat"%tracer), overwrite = True, format = 'ecsv')

    return HOD_dict



"""
The AbacusHOD module loads halo catalogs from the AbacusSummit 
simulations and outputs multi-tracer mock galaxy catalogs.  
The code is highly efficient and contains a large set of HOD
extensions such as secondary biases (assembly biases),
velocity biases, and satellite profile flexibilities. The baseline 
HODs are based on those from `Zheng et al. 2007 <https://arxiv.org/abs/astro-ph/0703457>`_ 
and `Alam et al. 2020 <http://arxiv.org/abs/1910.05095>`_. 
The HOD extensions are first explained in `Yuan et al. 2018 <https://arxiv.org/abs/1802.10115>`_, and more 
recently summarized in `Yuan et al. 2020b <https://arxiv.org/abs/2010.04182>`_ . 
This HOD code also supports RSD and incompleteness. The code is fast, 
completeling a :math:`(2Gpc/h)^3` volume in 80ms per tracer on a 32 core
desktop system, and the performance should be scalable. The module also
provides efficient correlation function and power spectrum calculators.
This module is particularly suited for efficiently sampling HOD parameter
space. We provide examples of docking it onto ``emcee`` and ``dynesty``
samplers. 

The module defines one class, ``AbacusHOD``, whose constructor 
takes the path to the simulation volume, and a set of HOD 
parameters, and runs the ``staging`` function to compile the 
simulation halo catalog as a set of arrays that are saved on
memory. The ``run_hod`` function can then be called to 
generate galaxy catalogs. 

The output takes the format of a dictionary of dictionaries,
where each subdictionary corresponds to a different tracer. 
Currently, we have enabled tracer types: LRG, ELG, and QSO.
Each subdictionary contains all the mock galaxies of that 
tracer type, recording their properties with keys ``x``, ``y``
, ``z``, ``vx``, ``vy``, ``vz``, ``mass``, ``id``, ``Ncent``.
The coordinates are in Mpc/h, and the velocities are in km/s.
The ``mass`` refers to host halo mass and is in units of Msun/h.
The ``id`` refers to halo id, and the ``Ncent`` key refers to number of
central galaxies for that tracer. The first ``Ncent`` galaxies 
in the catalog are always centrals and the rest are satellites. 

The galaxies can be written to disk by setting the 
``write_to_disk`` flag to ``True`` in the argument of 
``run_hod``. However, the I/O is slow and the``write_to_disk`` 
flag defaults to ``False``.

The core of the AbacusHOD code is a two-pass memory-in-place algorithm.
The first pass of the halo+particle subsample computes the number
of galaxies generated in total. Then an empty array for these galaxies 
is allocated in memory, which is then filled on the second pass of 
the halos+particles. Each pass is accelerated with numba parallel.
The default threading is set to 16. 


Theory
======
The baseline HOD for LRGs comes from Zheng et al. 2007:

.. math:: \\bar{n}_{\\mathrm{cent}}(M) = \\frac{1}{2}\\mathrm{erfc} \\left[\\frac{\\ln(M_{\\mathrm{cut}}/M)}{\\sqrt{2}\\sigma}\\right],
.. math:: \\bar{n}_{\\textrm{sat}}(M) = \\left[\\frac{M-\\kappa M_{\\textrm{cut}}}{M_1}\\right]^{\\alpha}\\bar{n}_{\\mathrm{cent}}(M),

The baseline HOD for ELGs and QSOs 
comes from Alam et al. 2020. The actual calculation
is complex and we refer the readers 
to section 3 of `said paper <http://arxiv.org/abs/1910.05095>`_ for details. 

In the baseline implementation, the central galaxy is assigned to the center 
of mass of the halo, with the velocity vector also set to that of the center 
of mass of the halo. Satellite galaxies are assigned to particles of the 
halo with equal weights. When multiple tracers are enabled, each halo/particle
can only host a single tracer type. However, we have not yet implemented any
prescription of conformity. 

The secondary bias (assembly bias) extensions follow the recipes described in 
`Xu et al. 2020 <https://arxiv.org/abs/2007.05545>`_ , where the secondary halo
property (concentration or local overdensity) is directly tied to the mass 
parameters in the baseline HOD (:math:`M_{\\mathrm{cut}}` and :math:`M_1`):

.. math:: \\log_{10} M_{\\mathrm{cut}}^{\\mathrm{mod}} = \\log_{10} M_{\\mathrm{cut}} + A_c(c^{\\mathrm{rank}} - 0.5) + B_c(\\delta^{\\mathrm{rank}} - 0.5)
.. math:: \\log_{10} M_{1}^{\\mathrm{mod}} = \\log_{10} M_{1} + A_s(c^{\\mathrm{rank}} - 0.5) + B_s(\\delta^{\\mathrm{rank}} - 0.5)

where :math:`c` and :math:`\\delta` represent the halo concentration and local 
overdensity, respectively. These secondary properties are ranked within narrow
halo mass bins, and the rank are normalized to range from 0 to 1, as noted by 
the :math:`\\mathrm{rank}` superscript. :math:`(A_c, B_c, A_s, B_s)` form the 
four parameters describing secondary biases in the HOD model. The default for
these parameters are 0. 

The velocity bias extension follows the common prescription as described in 
`Guo et al. 2015 <https://arxiv.org/abs/1407.4811>`_ . 

.. math:: \\sigma_c = \\alpha_c \\sigma_h
.. math:: v_s - v_h = \\alpha_s (v_p - v_h)

where the central velocity bias parameter :math:`\\alpha_c` sets the ratio of
central velocity dispersion vs. halo velocity dispersion. The satellite 
velocity bias parameter :math:`\\alpha_c` sets the ratio between the satellite
peculiar velocity to the particle peculiar velocity. The default for these two
parameters are 1. 

We additionaly introduce a set of satellite profile parameters 
:math:`(s, s_v, s_p, s_r)` that allow for flexibilities in how satellite 
galaxies are distributed within a halo. They respecctively allow the galaxy
weight per particle to depend on radial position (:math:`s`), peculair velocity
(:math:`s_v`), perihelion distance of the particle orbit (:math:`s_p`), and
the radial velocity (:math:`s_v`). The default values for these parameters are
0. A detailed description of these parameters are available in 
`Yuan et al. 2018 <https://arxiv.org/abs/1802.10115>`_, and more 
recently in `Yuan et al. 2020b <https://arxiv.org/abs/2010.04182>`_ . 


Some brief examples and technical details about the module
layout are presented below, followed by the full module API.


Short Example
=============

The first step is to create the configuration file such as ``config/abacus_hod.yaml``,
which provides the full customizability of the HOD code. By default, it lives in your 
current work directory under a subdirectory ``./config``. A template with 
default settings are provided under ``abacusutils/scripts/config``.

With the first use, you should define which simulation box, which redshift,
the path to simulation data, the path to output datasets, the various HOD 
flags and an initial set of HOD parameters. Other decisions that need to be 
made initially (you can always re-do this but it would take some time) include:
do you only want LRGs or do you want other tracers as well? 
Do you want to enable satellite profile flexibilities (the :math:`s, s_v, s_p, s_r`
parameters)? If so, you need to turn on ``want_ranks`` flag in the config file. 
If you want to enable secondary bias, you need to set ``want_AB`` flag to true in the
config file. The local environment is defined by total mass within 5 Mpc/h but beyond
``r98``. 

IMPORTANT: Running this code is a two-part process. First, you need to run the ``prepare_sim``
code, which generates the necessary data files for that simulation. Then you can run the actual
HOD code. The first step only needs to be done once for a simulation box, but it can be slow, 
depending on the downsampling and the features you choose to enable. 

So first, you need to run the ``prepare_sim`` script, this extracts the simulation outputs
and organizes them into formats that are suited for the HOD code. This code can take 
approximately an hour depending on your configuration settings and system capabilities. 
We recommend setting the ``Nthread_load`` parameter to ``min(sys_core_count, memoryGB_divided_by_30)``.
You can run ``load_sims`` on command line with ::

    python -m abacusnbody.hod.prepare_sim --path2config PATH2CONFIG


Within Python, you can run the same script with ``from abacusnbody.hod import prepare_sim``
and then ``prepare_sim.main(/path/to/config.yaml)``.

If your config file lives in the default location, i.e. ``./config``, then you 
can ignore the ``-path2config`` flag. 
Once that is finished, you can construct the ``AbacusHOD`` object and run fast 
HOD chains. A code template is given in ``abacusutils/scripts/run_hod.py`` for 
running a few example HODs and ``abacusutils/scripts/run_emcee.py`` for integrating 
with the ``emcee`` sampler. 

To use the given ``run_hod.py`` script to run a custom configuration file, you can
simply run the given script in bash ::

    python run_hod.py --path2config PATH2CONFIG

You can also consruct the AbacusHOD object yourself within Python and run HODs from
there. Here we show the scripts within ``run_hod.py`` for reference.::

    import os
    import glob
    import time
    import yaml
    import numpy as np
    import argparse

    from abacusnbody.hod.abacus_hod import AbacusHOD

    path2config = 'config/abacus_hod.yaml' # path to config file

    # load the config file and parse in relevant parameters
    config = yaml.safe_load(open(path2config))
    sim_params = config['sim_params']
    HOD_params = config['HOD_params']
    clustering_params = config['clustering_params']

    # additional parameter choices
    want_rsd = HOD_params['want_rsd']
    write_to_disk = HOD_params['write_to_disk']

    # create a new AbacusHOD object
    newBall = AbacusHOD(sim_params, HOD_params, clustering_params)
        
    # first hod run, slow due to compiling jit, write to disk
    mock_dict = newBall.run_hod(newBall.tracers, want_rsd, write_to_disk, Nthread = 16)

    # run the 10 different HODs for timing
    for i in range(10):
        newBall.tracers['LRG']['alpha'] += 0.01
        print("alpha = ",newBall.tracers['LRG']['alpha'])
        start = time.time()
        mock_dict = newBall.run_hod(newBall.tracers, want_rsd, write_to_disk = False, Nthread = 64)
        print("Done iteration ", i, "took time ", time.time() - start)

The class also provides fast 2PCF calculators. For example to compute the 
redshift-space 2PCF (:math:`\\xi(r_p, \\pi)`): ::

    # load the rp pi binning from the config file
    bin_params = clustering_params['bin_params']
    rpbins = np.logspace(bin_params['logmin'], bin_params['logmax'], bin_params['nbins'])
    pimax = clustering_params['pimax']
    pi_bin_size = clustering_params['pi_bin_size']    # the pi binning is configrured by pi_max and bin size

    mock_dict = newBall.run_hod(newBall.tracers, want_rsd, write_to_disk)
    xirppi = newBall.compute_xirppi(mock_dict, rpbins, pimax, pi_bin_size)

"""
import os
import glob
import time
import timeit
from pathlib import Path

import numpy as np
import h5py
import asdf
import argparse
import multiprocessing
from multiprocessing import Pool
from astropy.io import ascii

from .GRAND_HOD import *
from .parallel_numpy_rng import *
from .tpcf_corrfunc import calc_xirppi_fast, calc_wp_fast, calc_multipole_fast
# TODO B.H.: staging can be shorter and prettier; perhaps asdf for h5 and ecsv?

class AbacusHOD:
    """
    A highly efficient multi-tracer HOD code for the AbacusSummmit simulations.
    """
    def __init__(self, sim_params, HOD_params, clustering_params):
        """
        Loads simulation. The ``sim_params`` dictionary specifies which simulation
        volume to load. The ``HOD_params`` specifies the HOD parameters and tracer
        configurations. The ``clustering_params`` specifies the summary statistics 
        configurations. The ``HOD_params`` and ``clustering_params`` can be set to their
        default values in the ``config/abacus_hod.yaml`` file and changed later. 
        The ``sim_params`` cannot be changed once the ``AbacusHOD`` object is created. 

        Parameters
        ----------
        sim_params: dict
            Dictionary of simulation parameters. Load from ``config/abacus_hod.yaml``. The dictionary should contain the following keys:
                * ``sim_name``: str, name of the simulation volume, e.g. 'AbacusSummit_base_c000_ph006'. 
                * ``sim_dir``: str, the directory that the simulation lives in, e.g. '/path/to/AbacusSummit/'.                                 
                * ``output_dir``: str, the diretory to save galaxy to, e.g. '/my/output/galalxies'. 
                * ``subsample_dir``: str, where to save halo+particle subsample, e.g. '/my/output/subsamples/'. 
                * ``z_mock``: float, which redshift slice, e.g. 0.5.    

        HOD_params: dict 
            HOD parameters and tracer configurations. Load from ``config/abacus_hod.yaml``. It contains the following keys:
                * ``tracer_flags``: dict, which tracers is enabled: 
                    * ``LRG``: bool, default ``True``. 
                    * ``ELG``: bool, default ``False``. 
                    * ``QSO``: bool, default ``False``. 
                * ``want_ranks``: bool, enable satellite profile flexibilities. If ``False``, satellite profile follows the DM, default ``True``. 
                * ``want_rsd``: bool, enable RSD? default ``True``.                 # want RSD? 
                * ``Ndim``: int, grid density for computing local environment, default 1024.
                * ``density_sigma``: float, scale radius in Mpc / h for local density definition, default 3.
                * ``write_to_disk``: bool, output to disk? default ``False``. Setting to ``True`` decreases performance. 
                * ``LRG_params``: dict, HOD parameter values for LRGs. Default values are given in config file. 
                * ``ELG_params``: dict, HOD parameter values for ELGs. Default values are given in config file. 
                * ``QSO_params``: dict, HOD parameter values for QSOs. Default values are given in config file. 

        clustering_params: dict
            Summary statistics configuration parameters. Load from ``config/abacus_hod.yaml``. It contains the following keys:
                * ``clustering_type``: str, which summary statistic to compute. Options: ``wp``, ``xirppi``, default: ``xirppi``.
                * ``bin_params``: dict, transverse scale binning. 
                    * ``logmin``: float, :math:`\\log_{10}r_{\\mathrm{min}}` in Mpc/h.
                    * ``logmax``: float, :math:`\\log_{10}r_{\\mathrm{max}}` in Mpc/h.
                    * ``nbins``: int, number of bins.
                * ``pimax``: int, :math:`\\pi_{\\mathrm{max}}`. 
                * ``pi_bin_size``: int, size of bins along of the line of sight. Need to be divisor of ``pimax``.

        """
        # simulation details
        self.sim_name = sim_params['sim_name']
        self.sim_dir = sim_params['sim_dir']
        self.subsample_dir = sim_params['subsample_dir']
        self.z_mock = sim_params['z_mock']
        self.output_dir = sim_params['output_dir']
        
        # tracers
        tracer_flags = HOD_params['tracer_flags']
        tracers = {}
        for key in tracer_flags.keys():
            if tracer_flags[key]:
                tracers[key] = HOD_params[key+'_params']
        self.tracers = tracers

        # HOD parameter choices
        self.want_ranks = HOD_params['want_ranks']
        self.want_rsd = HOD_params['want_rsd']

        # clusteringparameters
        self.pimax = clustering_params['pimax']
        self.pi_bin_size = clustering_params['pi_bin_size']
        bin_params = clustering_params['bin_params']
        self.rpbins = np.logspace(bin_params['logmin'], bin_params['logmax'], bin_params['nbins'] + 1)
        self.clustering_type = clustering_params['clustering_type']
                
        # load the subsample particles
        self.halo_data, self.particle_data, self.params, self.mock_dir = self.staging()

        # determine the halo mass function
        self.logMbins = np.linspace(
            np.log10(np.min(self.halo_data['hmass'])), 
            np.log10(np.max(self.halo_data['hmass'])), 101)
        self.deltacbins = np.linspace(-0.5, 0.5, 101)
        self.fenvbins = np.linspace(-0.5, 0.5, 101)

        self.halo_mass_func, edges = np.histogramdd(
            np.vstack((np.log10(self.halo_data['hmass']), self.halo_data['hdeltac'], self.halo_data['hfenv'])).T,
            bins = [self.logMbins, self.deltacbins, self.fenvbins],
            weights = self.halo_data['hmultis'])


    def staging(self):
        """
        Constructor call this function to load the halo+particle subsamples onto memory. 
        """
        # all paths relevant for mock generation
        output_dir = Path(self.output_dir)
        simname = Path(self.sim_name)
        sim_dir = Path(self.sim_dir)
        mock_dir = output_dir / simname / ('z%4.3f'%self.z_mock)
        # create mock_dir if not created
        mock_dir.mkdir(parents = True, exist_ok = True)
        subsample_dir = \
            Path(self.subsample_dir) / simname / ('z%4.3f'%self.z_mock)

        # load header to read parameters
        halo_info_fns = \
            list((sim_dir / simname / 'halos' / ('z%4.3f'%self.z_mock) / 'halo_info').glob('*.asdf'))
        f = asdf.open(halo_info_fns[0], lazy_load=True, copy_arrays=False)
        header = f['header']

        # constants
        params = {}
        params['z'] = self.z_mock
        params['h'] = header['H0']/100.
        params['Lbox'] = header['BoxSize'] # Mpc / h, box size
        params['Mpart'] = header['ParticleMassHMsun']  # Msun / h, mass of each particle
        params['velz2kms'] = header['VelZSpace_to_kms']/params['Lbox']
        params['numslabs'] = len(halo_info_fns)
        self.lbox = header['BoxSize']

        # count ther number of halos and particles
        Nhalos = np.empty(params['numslabs'])
        Nparts = np.empty(params['numslabs'])        
        for eslab in range(params['numslabs']):
            
            if 'ELG' not in self.tracers.keys() and 'QSO' not in self.tracers.keys():
                halofilename = subsample_dir / ('halos_xcom_%d_seed600_abacushod_oldfenv'%eslab)
                particlefilename = subsample_dir / ('particles_xcom_%d_seed600_abacushod_oldfenv'%eslab)
            else:
                halofilename = subsample_dir / ('halos_xcom_%d_seed600_abacushod_oldfenv_MT'%eslab)
                particlefilename = subsample_dir / ('particles_xcom_%d_seed600_abacushod_oldfenv_MT'%eslab)            

            if self.want_ranks:
                particlefilename = str(particlefilename) + '_withranks'
            halofilename = str(halofilename) + '_new.h5'
            particlefilename = str(particlefilename) + '_new.h5'

            newfile = h5py.File(halofilename, 'r')
            newpart = h5py.File(particlefilename, 'r')
            Nhalos[eslab] = len(newfile['halos'])
            Nparts[eslab] = len(newpart['particles'])
        Nhalos = Nhalos.astype(int)
        Nparts = Nparts.astype(int)
        Nhalos_tot = int(np.sum(Nhalos))
        Nparts_tot = int(np.sum(Nparts))

        # list holding individual slabs
        hpos = np.empty((Nhalos_tot, 3))
        hvel = np.empty((Nhalos_tot, 3))
        hmass = np.empty([Nhalos_tot])
        hid = np.empty([Nhalos_tot], dtype = int)
        hmultis = np.empty([Nhalos_tot])
        hrandoms = np.empty([Nhalos_tot])
        hveldev = np.empty([Nhalos_tot])
        hsigma3d = np.empty([Nhalos_tot])
        hdeltac = np.empty([Nhalos_tot])
        hfenv = np.empty([Nhalos_tot])

        ppos = np.empty((Nparts_tot, 3))
        pvel = np.empty((Nparts_tot, 3))
        phvel = np.empty((Nparts_tot, 3))
        phmass = np.empty([Nparts_tot])
        phid = np.empty([Nparts_tot], dtype = int)
        pNp = np.empty([Nparts_tot])
        psubsampling = np.empty([Nparts_tot])
        prandoms = np.empty([Nparts_tot])
        pdeltac = np.empty([Nparts_tot])
        pfenv = np.empty([Nparts_tot])

        # ranks
        if self.want_ranks:
            p_ranks = np.empty([Nparts_tot])
            p_ranksv = np.empty([Nparts_tot])
            p_ranksp = np.empty([Nparts_tot])
            p_ranksr = np.empty([Nparts_tot])

        # B.H. make into ASDF
        # load all the halo and particle data we need
        halo_ticker = 0
        parts_ticker = 0
        for eslab in range(params['numslabs']):
            print("Loading simulation by slab, ", eslab)
            
            if 'ELG' not in self.tracers.keys() and 'QSO' not in self.tracers.keys():
                halofilename = subsample_dir / ('halos_xcom_%d_seed600_abacushod_oldfenv'%eslab)
                particlefilename = subsample_dir / ('particles_xcom_%d_seed600_abacushod_oldfenv'%eslab)
            else:
                halofilename = subsample_dir / ('halos_xcom_%d_seed600_abacushod_oldfenv_MT'%eslab)
                particlefilename = subsample_dir / ('particles_xcom_%d_seed600_abacushod_oldfenv_MT'%eslab)            

            if self.want_ranks:
                particlefilename = str(particlefilename) + '_withranks'
            halofilename = str(halofilename) + '_new.h5'
            particlefilename = str(particlefilename) + '_new.h5'

            newfile = h5py.File(halofilename, 'r')
            maskedhalos = newfile['halos']

            # extracting the halo properties that we need
            halo_ids = maskedhalos["id"].astype(int) # halo IDs
            halo_pos = maskedhalos["x_L2com"] # halo positions, Mpc / h
            halo_vels = maskedhalos['v_L2com'] # halo velocities, km/s
            halo_vel_dev = maskedhalos["randoms_gaus_vrms"] # halo velocity dispersions, km/s
            halo_sigma3d = maskedhalos["sigmav3d_L2com"] # 3d velocity dispersion
            halo_mass = maskedhalos['N']*params['Mpart'] # halo mass, Msun / h, 200b
            halo_deltac = maskedhalos['deltac_rank'] # halo concentration
            halo_fenv = maskedhalos['fenv_rank'] # halo velocities, km/s
            halo_pstart = maskedhalos['npstartA'].astype(int) # starting index of particles
            halo_pnum = maskedhalos['npoutA'].astype(int) # number of particles 
            halo_multi = maskedhalos['multi_halos']
            halo_submask = maskedhalos['mask_subsample'].astype(bool)
            halo_randoms = maskedhalos['randoms']

            hpos[halo_ticker: halo_ticker + Nhalos[eslab]] = halo_pos
            hvel[halo_ticker: halo_ticker + Nhalos[eslab]] = halo_vels
            hmass[halo_ticker: halo_ticker + Nhalos[eslab]] = halo_mass
            hid[halo_ticker: halo_ticker + Nhalos[eslab]] = halo_ids
            hmultis[halo_ticker: halo_ticker + Nhalos[eslab]] = halo_multi
            hrandoms[halo_ticker: halo_ticker + Nhalos[eslab]] = halo_randoms
            hveldev[halo_ticker: halo_ticker + Nhalos[eslab]] = halo_vel_dev
            hsigma3d[halo_ticker: halo_ticker + Nhalos[eslab]] = halo_sigma3d
            hdeltac[halo_ticker: halo_ticker + Nhalos[eslab]] = halo_deltac
            hfenv[halo_ticker: halo_ticker + Nhalos[eslab]] = halo_fenv
            halo_ticker += Nhalos[eslab]

            # extract particle data that we need
            newpart = h5py.File(particlefilename, 'r')
            subsample = newpart['particles']
            part_pos = subsample['pos']
            part_vel = subsample['vel']
            part_hvel = subsample['halo_vel']
            part_halomass = subsample['halo_mass'] # msun / h
            part_haloid = subsample['halo_id'].astype(int)
            part_Np = subsample['Np'] # number of particles that end up in the halo
            part_subsample = subsample['downsample_halo']
            part_randoms = subsample['randoms']
            part_deltac = subsample['halo_deltac']
            part_fenv = subsample['halo_fenv']

            if self.want_ranks:
                part_ranks = subsample['ranks']
                part_ranksv = subsample['ranksv']
                part_ranksp = subsample['ranksp']
                part_ranksr = subsample['ranksr']
                p_ranks[parts_ticker: parts_ticker + Nparts[eslab]] = part_ranks
                p_ranksv[parts_ticker: parts_ticker + Nparts[eslab]] = part_ranksv
                p_ranksp[parts_ticker: parts_ticker + Nparts[eslab]] = part_ranksp
                p_ranksr[parts_ticker: parts_ticker + Nparts[eslab]] = part_ranksr

            # #     part_data_slab += [part_ranks, part_ranksv, part_ranksp, part_ranksr]
            # particle_data = vstack([particle_data, new_part_table])
            ppos[parts_ticker: parts_ticker + Nparts[eslab]] = part_pos
            pvel[parts_ticker: parts_ticker + Nparts[eslab]] = part_vel
            phvel[parts_ticker: parts_ticker + Nparts[eslab]] =  part_hvel
            phmass[parts_ticker: parts_ticker + Nparts[eslab]] = part_halomass
            phid[parts_ticker: parts_ticker + Nparts[eslab]] =  part_haloid
            pNp[parts_ticker: parts_ticker + Nparts[eslab]] = part_Np
            psubsampling[parts_ticker: parts_ticker + Nparts[eslab]] = part_subsample
            prandoms[parts_ticker: parts_ticker + Nparts[eslab]] = part_randoms
            pdeltac[parts_ticker: parts_ticker + Nparts[eslab]] = part_deltac
            pfenv[parts_ticker: parts_ticker + Nparts[eslab]] = part_fenv
            parts_ticker += Nparts[eslab]

        halo_data = {"hpos": hpos, 
                     "hvel": hvel, 
                     "hmass": hmass, 
                     "hid": hid, 
                     "hmultis": hmultis, 
                     "hrandoms": hrandoms, 
                     "hveldev": hveldev, 
                     "hsigma3d": hsigma3d, 
                     "hdeltac": hdeltac, 
                     "hfenv": hfenv}
        pweights = 1/pNp/psubsampling
        particle_data = {"ppos": ppos, 
                         "pvel": pvel, 
                         "phvel": phvel, 
                         "phmass": phmass, 
                         "phid": phid, 
                         "pweights": pweights, 
                         "prandoms": prandoms, 
                         "pdeltac": pdeltac, 
                         "pfenv": pfenv}
        if self.want_ranks:
            particle_data['pranks'] = p_ranks
            particle_data['pranksv'] = p_ranksv
            particle_data['pranksp'] = p_ranksp
            particle_data['pranksr'] = p_ranksr
        else:
            particle_data['pranks'] = np.ones(Nparts_tot)
            particle_data['pranksv'] =  np.ones(Nparts_tot)
            particle_data['pranksp'] =  np.ones(Nparts_tot)
            particle_data['pranksr'] =  np.ones(Nparts_tot)
        
        return halo_data, particle_data, params, mock_dir

    def run_hod(self, tracers = None, want_rsd = True, reseed = None, write_to_disk = False, 
        Nthread = 16, verbose = False):
        """
        Runs a custom HOD.

        Parameters
        ----------
        ``tracers``: dict
            dictionary of multi-tracer HOD. ``tracers['LRG']`` is the dictionary of LRG HOD parameters,
            overwrites the ``LRG_params`` argument in the constructor.
            Same for keys ``'ELG'`` and ``'QSO'``.
        
        ``want_rsd``: bool 
            enable RSD? default ``True``.
            
        ``reseed``: int
            re-generate random numbers? supply random number seed. This overwrites the pre-generated random numbers, at a performance cost.
            Default ``None``.

        ``write_to_disk``: bool 
            output to disk? default ``False``. Setting to ``True`` decreases performance. 

        ``Nthread``: int
            number of threads in the HOD run. Default 16. 

        ``verbose``: bool, 
            detailed stdout? default ``False``.

        Returns
        -------
        mock_dict: dict
            dictionary of galaxy outputs. Contains keys ``'LRG'``, ``'ELG'``, and ``'QSO'``. Each
            tracer key corresponds to a sub-dictionary that contains the galaxy properties with keys 
            ``'x'``, ``'y'``, ``'z'``, ``'vx'``, ``'vy'``, ``'vz'``, ``'mass'``, ``'id'``, ``Ncent'``.
            The coordinates are in Mpc/h, and the velocities are in km/s.
            The ``'mass'`` refers to host halo mass and is in units of Msun/h.
            The ``'id'`` refers to halo id, and the ``'Ncent'`` key refers to number of
            central galaxies for that tracer. The first ``'Ncent'`` galaxies 
            in the catalog are always centrals and the rest are satellites. 

        """
        if tracers == None:
            tracers = self.tracers
        if reseed:
            start = time.time()
            mtg = MTGenerator(np.random.PCG64(reseed))
            r1 = mtg.random(size=len(self.halo_data['hrandoms']), nthread=Nthread, dtype=np.float32)
            r2 = mtg.standard_normal(size=len(self.halo_data['hveldev']), nthread=Nthread, dtype=np.float32)*self.halo_data['hsigma3d']/np.sqrt(3)
            r3 = mtg.random(size=len(self.particle_data['prandoms']), nthread=Nthread, dtype=np.float32)
            self.halo_data['hrandoms'] = r1
            self.halo_data['hveldev'] = r2
            self.particle_data['prandoms'] = r3
            
            print("gen randoms took, ", time.time() - start)
            
        start = time.time()
        mock_dict = gen_gal_cat(self.halo_data, self.particle_data, tracers, self.params, Nthread, 
            enable_ranks = self.want_ranks, 
            rsd = want_rsd, 
            write_to_disk = write_to_disk, 
            savedir = self.mock_dir,
            verbose = False)
        print("gen mocks", time.time() - start)

        return mock_dict

    def compute_ngal(self, tracers = None, Nthread = 16):
        """
        Computes the number of each tracer generated by the HOD

        Parameters
        ----------
        ``tracers``: dict
            dictionary of multi-tracer HOD. ``tracers['LRG']`` is the dictionary of LRG HOD parameters,
            overwrites the ``LRG_params`` argument in the constructor.
            Same for keys ``'ELG'`` and ``'QSO'``.

        ``Nthread``: int
            Number of threads in the HOD run. Default 16. 

        Returns
        -------
        ngal_dict: dict
        dictionary of number of each tracer. 

        fsat_dict: dict
        dictionary of satellite fraction of each tracer.

        """
        if tracers == None:
            tracers = self.tracers

        ngal_dict = {}
        fsat_dict = {}
        for etracer in tracers.keys():
            tracer_hod = tracers[etracer]
            if etracer == 'LRG':
                newngal = AbacusHOD._compute_ngal_lrg(
                    self.logMbins, self.deltacbins, self.fenvbins, self.halo_mass_func,
                    tracer_hod['logM_cut'], tracer_hod['logM1'], tracer_hod['sigma'], 
                    tracer_hod['alpha'], tracer_hod['kappa'], tracer_hod['Acent'], 
                    tracer_hod['Asat'], tracer_hod['Bcent'], tracer_hod['Bsat'], tracer_hod['ic'], Nthread)
                ngal_dict[etracer] = newngal[0] + newngal[1]
                fsat_dict[etracer] = newngal[1] / (newngal[0] + newngal[1])
            elif etracer == 'ELG':
                newngal = AbacusHOD._compute_ngal_elg(
                    self.logMbins, self.deltacbins, self.fenvbins, self.halo_mass_func, 
                    tracer_hod['p_max'], tracer_hod['Q'], tracer_hod['logM_cut'], 
                    tracer_hod['kappa'], tracer_hod['sigma'], tracer_hod['logM1'],  
                    tracer_hod['alpha'], tracer_hod['gamma'], tracer_hod['A_s'],  
                    tracer_hod['Acent'],  tracer_hod['Asat'],  tracer_hod['Bcent'],  tracer_hod['Bsat'], tracer_hod['ic'], Nthread) 
                ngal_dict[etracer] = newngal[0] + newngal[1]
                fsat_dict[etracer] = newngal[1] / (newngal[0] + newngal[1])
            elif etracer == 'QSO':
                newngal = AbacusHOD._compute_ngal_qso(
                    self.logMbins, self.deltacbins, self.fenvbins, self.halo_mass_func, 
                    tracer_hod['p_max'], tracer_hod['logM_cut'], 
                    tracer_hod['kappa'], tracer_hod['sigma'], tracer_hod['logM1'],  
                    tracer_hod['alpha'], tracer_hod['A_s'],  
                    tracer_hod['Acent'],  tracer_hod['Asat'],  tracer_hod['Bcent'],  tracer_hod['Bsat'], tracer_hod['ic'], Nthread)         
                ngal_dict[etracer] = newngal[0] + newngal[1]
                fsat_dict[etracer] = newngal[1] / (newngal[0] + newngal[1])
        return ngal_dict, fsat_dict

    @staticmethod
    @njit(fastmath = True, parallel = True)
    def _compute_ngal_lrg(logMbins, deltacbins, fenvbins, halo_mass_func,
                   logM_cut, logM1, sigma, alpha, kappa, Acent, Asat, Bcent, Bsat, ic, Nthread):
        """
        internal helper to compute number of LRGs
        """
        numba.set_num_threads(Nthread)

        logMs = 0.5*(logMbins[1:] + logMbins[:-1])
        deltacs = 0.5*(deltacbins[1:] + deltacbins[:-1])
        fenvs = 0.5*(fenvbins[1:] + fenvbins[:-1])
        ngal_cent = 0
        ngal_sat = 0
        for i in numba.prange(len(logMbins) - 1):
            for j in range(len(deltacbins) - 1):
                for k in range(len(fenvbins) - 1):
                    Mh_temp = 10**logMs[i]
                    logM_cut_temp = logM_cut + Acent * deltacs[j] + Bcent * fenvs[k]
                    M1_temp = 10**(logM1 + Asat * deltacs[j] + Bsat * fenvs[k])
                    ncent_temp = n_cen_LRG(Mh_temp, logM_cut_temp, sigma)
                    nsat_temp = n_sat_LRG_modified(Mh_temp, logM_cut_temp, 
                        10**logM_cut_temp, M1_temp, sigma, alpha, kappa)
                    ngal_cent += halo_mass_func[i, j, k] * ncent_temp * ic
                    ngal_sat += halo_mass_func[i, j, k] * nsat_temp * ic
        return ngal_cent, ngal_sat

    @staticmethod
    @njit(fastmath = True, parallel = True)
    def _compute_ngal_elg(logMbins, deltacbins, fenvbins, halo_mass_func, p_max, Q,
                   logM_cut, kappa, sigma, logM1, alpha, gamma, A_s, Acent, Asat, Bcent, Bsat, ic, Nthread):
        """
        internal helper to compute number of LRGs
        """
        numba.set_num_threads(Nthread)

        logMs = 0.5*(logMbins[1:] + logMbins[:-1])
        deltacs = 0.5*(deltacbins[1:] + deltacbins[:-1])
        fenvs = 0.5*(fenvbins[1:] + fenvbins[:-1])
        ngal_cent = 0
        ngal_sat = 0
        for i in numba.prange(len(logMbins) - 1):
            for j in range(len(deltacbins) - 1):
                for k in range(len(fenvbins) - 1):
                    Mh_temp = 10**logMs[i]
                    logM_cut_temp = logM_cut + Acent * deltacs[j] + Bcent * fenvs[k]
                    M1_temp = 10**(logM1 + Asat * deltacs[j] + Bsat * fenvs[k])
                    ncent_temp = N_cen_ELG_v1(Mh_temp, p_max, Q, logM_cut_temp, sigma, gamma, A_s)
                    nsat_temp = N_sat_generic(Mh_temp, 10**logM_cut_temp, kappa, M1_temp, alpha)
                    ngal_cent += halo_mass_func[i, j, k] * ncent_temp * ic
                    ngal_sat += halo_mass_func[i, j, k] * nsat_temp * ic
        return ngal_cent, ngal_sat

    @staticmethod
    @njit(fastmath = True, parallel = True)
    def _compute_ngal_qso(logMbins, deltacbins, fenvbins, halo_mass_func, p_max,
                   logM_cut, kappa, sigma, logM1, alpha, A_s, Acent, Asat, Bcent, Bsat, ic, Nthread):
        """
        internal helper to compute number of LRGs
        """
        numba.set_num_threads(Nthread)

        logMs = 0.5*(logMbins[1:] + logMbins[:-1])
        deltacs = 0.5*(deltacbins[1:] + deltacbins[:-1])
        fenvs = 0.5*(fenvbins[1:] + fenvbins[:-1])
        ngal_cent = 0
        ngal_sat = 0
        for i in numba.prange(len(logMbins) - 1):
            for j in range(len(deltacbins) - 1):
                for k in range(len(fenvbins) - 1):
                    Mh_temp = 10**logMs[i]
                    logM_cut_temp = logM_cut + Acent * deltacs[j] + Bcent * fenvs[k]
                    M1_temp = 10**(logM1 + Asat * deltacs[j] + Bsat * fenvs[k])
                    ncent_temp = N_cen_QSO(Mh_temp, p_max, logM_cut_temp, sigma)
                    nsat_temp = N_sat_generic(Mh_temp, 10**logM_cut_temp, kappa, M1_temp, alpha, A_s)
                    ngal_cent += halo_mass_func[i, j, k] * ncent_temp * ic
                    ngal_sat += halo_mass_func[i, j, k] * nsat_temp * ic
        return ngal_cent, ngal_sat

    def compute_clustering(self, mock_dict, *args, **kwargs):
        """
        Computes summary statistics, currently enabling ``wp`` and ``xirppi``.

        Parameters
        ----------
        ``mock_dict``: dict
            dictionary of tracer positions. Output of ``run_hod``. 

        ``Ntread``: int
            number of threads in the HOD run. Default 16. 

        ``rpbins``: np.array
            array of transverse bins in Mpc/h.

        ``pimax``: int
            maximum bin edge along the line of sight direction, in Mpc/h. 

        ``pi_bin_size``: int
            size of bin along the line of sight. Currently, we only support linear binning along the line of sight. 

        Returns
        -------
        clustering: dict
            dicionary of summary statistics. Auto-correlations/spectra can be
            accessed with keys such as ``'LRG_LRG'``. Cross-correlations/spectra can be 
            accessed with keys such as ``'LRG_ELG'``. 
        """
        if self.clustering_type == 'xirppi':
            clustering = self.compute_xirppi(mock_dict, *args, **kwargs)
        elif self.clustering_type == 'wp':
            clustering = self.compute_wp(mock_dict, *args, **kwargs)
        elif self.clustering_type == 'multipole':
            clustering = self.compute_multipole(mock_dict, *args, **kwargs)
        return clustering
    
    def compute_xirppi(self, mock_dict, rpbins, pimax, pi_bin_size, Nthread = 8):
        """
        Computes :math:`\\xi(r_p, \\pi)`.

        Parameters
        ----------
        ``mock_dict``: dict
            dictionary of tracer positions. Output of ``run_hod``. 

        ``Ntread``: int
            number of threads in the HOD run. Default 16. 

        ``rpbins``: np.array
            array of transverse bins in Mpc/h.

        ``pimax``: int
            maximum bin edge along the line of sight direction, in Mpc/h. 

        ``pi_bin_size``: int
            size of bin along the line of sight. Currently, we only support linear binning along the line of sight. 

        Returns
        -------
        clustering: dict
            dicionary of summary statistics. Auto-correlations/spectra can be
            accessed with keys such as ``'LRG_LRG'``. Cross-correlations/spectra can be 
            accessed with keys such as ``'LRG_ELG'``. 
        """
        clustering = {}
        for i1, tr1 in enumerate(mock_dict.keys()):
            x1 = mock_dict[tr1]['x']
            y1 = mock_dict[tr1]['y']
            z1 = mock_dict[tr1]['z']
            for i2, tr2 in enumerate(mock_dict.keys()):
                if i1 > i2: continue # cross-correlations are symmetric
                if i1 == i2: # auto corr
                    clustering[tr1+'_'+tr2] = calc_xirppi_fast(x1, y1, z1, rpbins, pimax, pi_bin_size, 
                        self.lbox, Nthread)
                else:
                    x2 = mock_dict[tr2]['x']
                    y2 = mock_dict[tr2]['y']
                    z2 = mock_dict[tr2]['z']
                    clustering[tr1+'_'+tr2] = calc_xirppi_fast(x1, y1, z1, rpbins, pimax, pi_bin_size, 
                        self.lbox, Nthread, x2 = x2, y2 = y2, z2 = z2)
                    clustering[tr2+'_'+tr1] = clustering[tr1+'_'+tr2]
        return clustering

    def compute_multipole(self, mock_dict, rpbins, pimax, Nthread = 8):
        clustering = {}
        for i1, tr1 in enumerate(mock_dict.keys()):
            x1 = mock_dict[tr1]['x']
            y1 = mock_dict[tr1]['y']
            z1 = mock_dict[tr1]['z']
            for i2, tr2 in enumerate(mock_dict.keys()):
                if i1 > i2: continue # cross-correlations are symmetric
                if i1 == i2: # auto corr
                    new_multi = calc_multipole_fast(x1, y1, z1, rpbins,
                        self.lbox, Nthread)
                    new_wp = calc_wp_fast(x1, y1, z1, rpbins, pimax, self.lbox, Nthread)
                    clustering[tr1+'_'+tr2] = np.concatenate((new_wp, new_multi))
                else:
                    x2 = mock_dict[tr2]['x']
                    y2 = mock_dict[tr2]['y']
                    z2 = mock_dict[tr2]['z']
                    new_multi = calc_multipole_fast(x1, y1, z1, rpbins,
                        self.lbox, Nthread, x2 = x2, y2 = y2, z2 = z2)
                    new_wp = calc_wp_fast(x1, y1, z1, rpbins, pimax, self.lbox, Nthread, 
                        x2 = x2, y2 = y2, z2 = z2)
                    clustering[tr1+'_'+tr2] = np.concatenate((new_wp, new_multi))
                    clustering[tr2+'_'+tr1] = clustering[tr1+'_'+tr2]
        return clustering

    def compute_wp(self, mock_dict, rpbins, pimax, pi_bin_size, Nthread = 8):
        """
        Computes :math:`w_p`.

        Parameters
        ----------
        ``mock_dict``: dict
            dictionary of tracer positions. Output of ``run_hod``. 

        ``Ntread``: int
            number of threads in the HOD run. Default 16. 

        ``rpbins``: np.array
            array of transverse bins in Mpc/h.

        ``pimax``: int
            maximum bin edge along the line of sight direction, in Mpc/h. 

        ``pi_bin_size``: int
            size of bin along the line of sight. Currently, we only support linear binning along the line of sight. 

        Returns
        -------
        clustering: dict
            dicionary of summary statistics. Auto-correlations/spectra can be
            accessed with keys such as ``'LRG_LRG'``. Cross-correlations/spectra can be 
            accessed with keys such as ``'LRG_ELG'``. 
        """
        clustering = {}
        for i1, tr1 in enumerate(mock_dict.keys()):
            x1 = mock_dict[tr1]['x']
            y1 = mock_dict[tr1]['y']
            z1 = mock_dict[tr1]['z']
            for i2, tr2 in enumerate(mock_dict.keys()):
                if i1 > i2: continue # cross-correlations are symmetric
                if i1 == i2:
                    print(tr1+'_'+tr2)
                    clustering[tr1+'_'+tr2] = calc_wp_fast(x1, y1, z1, rpbins, pimax, self.lbox, Nthread)
                else:
                    print(tr1+'_'+tr2)
                    x2 = mock_dict[tr2]['x']
                    y2 = mock_dict[tr2]['y']
                    z2 = mock_dict[tr2]['z']
                    clustering[tr1+'_'+tr2] = calc_wp_fast(x1, y1, z1, rpbins, pimax, self.lbox, Nthread, 
                        x2 = x2, y2 = y2, z2 = z2)
                    clustering[tr2+'_'+tr1] = clustering[tr1+'_'+tr2]
        return clustering

    def gal_reader(self, output_dir = None, simname = None, 
        sim_dir = None, z_mock = None, want_rsd = None, tracers = None):
        """
        Loads galaxy data given directory and return a ``mock_dict`` dictionary.  

        Parameters
        ----------
        ``sim_name``: str
            name of the simulation volume, e.g. 'AbacusSummit_base_c000_ph006'. 

        ``sim_dir``: str
            the directory that the simulation lives in, e.g. '/path/to/AbacusSummit/'.                                 

        ``output_dir``: str
            the diretory to save galaxy to, e.g. '/my/output/galalxies'. 

        ``z_mock``: floa
             which redshift slice, e.g. 0.5.    

        ``want_rsd``: bool
            RSD?

        ``tracers``: dict
            dictionary of tracer types to load, e.g. `{'LRG', 'ELG'}`.

        Returns
        -------
        ``mock_dict``: dict
            dictionary of tracer positions. Output of ``run_hod``. 

        """

        if output_dir == None:
            output_dir = Path(self.output_dir)
        if simname == None:
            simname = Path(self.sim_name)
        if sim_dir == None:
            sim_dir = Path(self.sim_dir)
        if z_mock == None:
            z_mock = self.z_mock
        if want_rsd == None:
            want_rsd = self.want_rsd
        if tracers == None:
            tracers = self.tracers.keys()
        mock_dir = output_dir / simname / ('z%4.3f'%self.z_mock)

        if want_rsd:
            rsd_string = "_rsd"
        else:
            rsd_string = ""

        outdir = (self.mock_dir) / ("galaxies"+rsd_string)

        mockdict = {}
        for tracer in tracers:
            mockdict[tracer] = ascii.read(outdir/(tracer+'s.dat'))
        return mockdict





#!/usr/bin/env python2
import os, sys
from kvsclient import KVSClient

# Connect to the key value store service (its location is the first
# argument).
kvs = KVSClient(sys.argv[1] if len(sys.argv) > 1 else None)

# Find out which process we are (assumes running under SLURM).
rank = int(os.environ['SLURM_PROCID'])

# Example of a function that dynamically and disjointly decomposes the
# work. All participants make the same sequence of calls to this
# function. One and only one participant will see True on the Nth
# invocation, the rest will see False.
#
# "chunk" controls the coarseness of the decomposition.
def own(rank, chunk=1):
    # initialize a global counter.
    if 0 == rank: kvs.put('count', 0)
    checks, lower, upper = 0, -1, -1
    while 1:
        if checks >= upper:
            lower = kvs.get('count')
            upper = lower + chunk
            kvs.put('count', upper)
        yield lower <= checks < upper
        checks += 1

# Here we illustrate splitting a simple loop, but the same approach
# would work with any iterative control structure, as long as it is
# deterministic.
ownCheck = own(rank, chunk=300013)
s = 0
limit = 100000000
for x in xrange(limit):
    if ownCheck.next(): s += 2*x + 1
# Report partial result.
kvs.put('psum', (rank, s))

# One participant gathers the partial results and generates the final
# output.
if 0 == rank:
    s, workers = 0, int(os.environ['SLURM_NTASKS'])
    for p in xrange(workers):
        wrank, ps = kvs.get('psum')
        print('rank %d: %d'%(wrank, ps))
        s += ps
    print('total = %d (%r)' % (s, (limit*limit) == s))
    sys.exit((limit*limit) != s)



# This library can be used to submit tasks to disBatch.py --taskserver or --taskcommand

import kvsstcp, os

class DisBatcher(object):
    def __init__(self, tasksname='DisBatcher', kvsserver=None):
        self.kvs = kvsstcp.KVSClient(kvsserver)
        self.kvs.put('task source name', tasksname, False)
        self.taskkey = tasksname + ' task'
        self.donetask = tasksname + ' done!'
        self.resultkey = tasksname + ' result %d'
        self.taskCount = 0
        self.tx2rc = {}

    def done(self):
        self.kvs.put(self.taskkey, self.donetask, False)

    def submit(self, c):
        '''Add a task to the disBatch queue, returning its streamIndex (which may map to multiple taskIds).'''
        self.kvs.put(self.taskkey, c, False)
        self.taskCount += 1
        return self.taskCount

    def syncTasks(self, taskd):
        '''Wait for specified taskIds (not streamIndex) to complete and collect results, returning a dict from taskd keys to returncodes.'''
        tx2rc = {}
        for tx in taskd:
            if tx not in self.tx2rc:
                r = self.kvs.get(self.resultkey%tx, False)
                lags, taskId, streamIndex, repIndex, host, pid, returncode, time, start, end, outbytes, errbytes, cmd = r.split('\t', 12)
                # do something with the rest of these results?
                self.tx2rc[tx] = int(returncode)
            tx2rc[tx] = self.tx2rc[tx]
        return tx2rc



import os

from setuptools import setup, find_namespace_packages

install_requires = ['numpy>=1.16','blosc>=1.9.2','astropy>=4.0.0','scipy','numba>=0.50','asdf>=2.8','h5py','emcee','schwimmbad','getdist','dynesty','sklearn']

# If we're on ReadTheDocs, can't install packages with C dependencies, like Corrfunc
on_rtd = os.environ.get('READTHEDOCS') == 'True'
if not on_rtd:
    # This list must be kept synchronized with `autodoc_mock_imports` in docs/conf.py
    install_requires += ['Corrfunc>=2']

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="abacusutils",
    author="Daniel Eisenstein, Philip Pinto, Lehman Garrison, Nina Maksimova, Sownak Bose, Boryana Hadzhiyska, Sihan (Sandy) Yuan",
    author_email="lgarrison@flatironinstitute.org",
    description="Python and C/C++ code to read halo catalogs and other Abacus N-body data products",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/abacusorg/abacusutils",
    packages=find_namespace_packages(include=['abacusnbody.*']),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    ],
    python_requires='>=3.6',
    install_requires = install_requires,
    entry_points={'console_scripts':['pipe_asdf = abacusnbody.data.pipe_asdf:main'],
                  'asdf.extensions':['abacusutils = abacusnbody.data.asdf:AbacusExtension']}
)



"""
to test the hod code against reference, run 
    $ pytest tests/test_hod.py
from abacusutils/ 

to generate new reference, run 
    $ python tests/test_hod.py
from abacusutils/

"""

import tempfile
import filecmp
import os.path

import yaml
import pytest
import h5py
import numpy as np
from astropy.io import ascii

EXAMPLE_SIM = os.path.join(os.path.dirname(__file__), 'Mini_N64_L32')
EXAMPLE_CONFIG = os.path.join(os.path.dirname(__file__), 'abacus_hod.yaml')
EXAMPLE_SUBSAMPLE_HALOS = os.path.join(os.path.dirname(__file__), 
    'data_summit/Mini_N64_L32/z0.000/halos_xcom_2_seed600_abacushod_oldfenv_MT_ref.h5')
EXAMPLE_SUBSAMPLE_PARTS = os.path.join(os.path.dirname(__file__), 
    'data_summit/Mini_N64_L32/z0.000/particles_xcom_2_seed600_abacushod_oldfenv_MT_ref.h5')
EXAMPLE_LRGS = os.path.join(os.path.dirname(__file__), 
    'data_mocks_summit_new/Mini_N64_L32/z0.000/galaxies_rsd/LRGs.dat')
EXAMPLE_ELGS = os.path.join(os.path.dirname(__file__), 
    'data_mocks_summit_new/Mini_N64_L32/z0.000/galaxies_rsd/ELGs.dat')
path2config = os.path.join(os.path.dirname(__file__), 'abacus_hod.yaml')

# @pytest.mark.xfail
def test_hod(tmp_path, reference_mode = False):
    '''Test loading a halo catalog
    '''
    from abacusnbody.hod import prepare_sim
    from abacusnbody.hod.abacus_hod import AbacusHOD

    # reference mode
    if reference_mode:
        prepare_sim.main(path2config)

        # load the yaml parameters
        config = yaml.safe_load(open(path2config))
        sim_params = config['sim_params']
        HOD_params = config['HOD_params']
        clustering_params = config['clustering_params']
        
        # additional parameter choices
        want_rsd = HOD_params['want_rsd']
        write_to_disk = HOD_params['write_to_disk']
        bin_params = clustering_params['bin_params']
        rpbins = np.logspace(bin_params['logmin'], bin_params['logmax'], bin_params['nbins'])
        pimax = clustering_params['pimax']
        pi_bin_size = clustering_params['pi_bin_size']
        
        # create a new abacushod object
        newBall = AbacusHOD(sim_params, HOD_params, clustering_params)
        mock_dict = newBall.run_hod(newBall.tracers, want_rsd, write_to_disk = True, Nthread = 2)

    # test mode
    else:
        config = yaml.safe_load(open(EXAMPLE_CONFIG))
        sim_params = config['sim_params']
        HOD_params = config['HOD_params']
        clustering_params = config['clustering_params']

        simname = config['sim_params']['sim_name'] # "AbacusSummit_base_c000_ph006"
        simdir = config['sim_params']['sim_dir']
        z_mock = config['sim_params']['z_mock']
        config['sim_params']['subsample_dir'] = str(tmp_path) + "/data_subs/"
        config['sim_params']['scratch_dir'] = str(tmp_path) + "/data_gals/"
        savedir = config['sim_params']['subsample_dir'] + simname+"/z"+str(z_mock).ljust(5, '0')

        # check subsample file match
        prepare_sim.main(EXAMPLE_CONFIG, params = config)

        newhalos = h5py.File(savedir+'/halos_xcom_2_seed600_abacushod_oldfenv_MT_new.h5', 'r')['halos']
        temphalos = h5py.File(EXAMPLE_SUBSAMPLE_HALOS, 'r')['halos']
        for i in range(len(newhalos)):
            for j in range(len(newhalos[i])):
                assert np.array_equal(newhalos[i][j], temphalos[i][j])
        newparticles = h5py.File(savedir+'/particles_xcom_2_seed600_abacushod_oldfenv_MT_new.h5', 'r')['particles']
        tempparticles = h5py.File(EXAMPLE_SUBSAMPLE_PARTS, 'r')['particles']
        for i in range(len(newparticles)):
            for j in range(len(newparticles[i])):
                assert np.array_equal(newparticles[i][j], tempparticles[i][j])

        # additional parameter choices
        want_rsd = HOD_params['want_rsd']
        write_to_disk = HOD_params['write_to_disk']
        bin_params = clustering_params['bin_params']
        rpbins = np.logspace(bin_params['logmin'], bin_params['logmax'], bin_params['nbins'])
        pimax = clustering_params['pimax']
        pi_bin_size = clustering_params['pi_bin_size']
        
        # create a new abacushod object
        newBall = AbacusHOD(sim_params, HOD_params, clustering_params)
        
        # throw away run for jit to compile, write to disk
        mock_dict = newBall.run_hod(newBall.tracers, want_rsd, write_to_disk = True, Nthread = 2)
        savedir_gal = config['sim_params']['output_dir']\
            +"/"+simname+"/z"+str(z_mock).ljust(5, '0') +"/galaxies_rsd/LRGs.dat"
        data = ascii.read(EXAMPLE_LRGS)
        data1 = ascii.read(savedir_gal)
        for ekey in data.keys():
            assert np.allclose(data[ekey], data1[ekey])

        savedir_gal = config['sim_params']['output_dir']\
            +"/"+simname+"/z"+str(z_mock).ljust(5, '0') +"/galaxies_rsd/ELGs.dat"
        data = ascii.read(EXAMPLE_ELGS)
        data1 = ascii.read(savedir_gal)
        for ekey in data.keys():
            assert np.allclose(data[ekey], data1[ekey])

if __name__ == '__main__':
    test_hod(".", reference_mode = False)



'''
A collection of routines related to various Abacus bitpacked
formats, like RVint and encoding of information in the PIDs.

Most users will not use this module directly, but will instead use
:mod:`abacusnbody.data.compaso_halo_catalog` or
:func:`abacusnbody.data.read_abacus.read_asdf`.
'''

import numpy as np
import numba as nb

__all__ = ['unpack_rvint', 'unpack_pids']

# Constants
AUXDENS = 0x07fe000000000000
ZERODEN = 49 #The density bits are 49-58. 

AUXXPID = 0x7fff #bits 0-14
AUXYPID = 0x7fff0000 #bits 16-30
AUXZPID = 0x7fff00000000 #bits 32-46
AUXPID = AUXXPID | AUXYPID | AUXZPID # all of the above bits
AUXTAGGED = 48 # tagged bit is 48

# The names of the bit-packed PID fields that the user can request
PID_FIELDS=['pid', 'lagr_pos', 'tagged', 'density', 'lagr_idx']


def unpack_rvint(intdata, boxsize, float_dtype=np.float32, posout=None, velout=None):
    '''
    Unpack rvint data into pos and vel.
    
    Parameters
    ----------
    intdata: ndarray of dtype np.int32
        The rvint data
    boxsize: float
        The box size, used to scale the positions
    float_dtype: np.dtype, optional
        The precision in which to store the unpacked values.
        Default: np.float32
    posout: ndarray, None, or False; optional
        The array in which to store the unpacked positions.
        `None` can be given (the default), in which case an
        array is constructed and retured as the first return value.
        `False` can be given, in which case the positions are not unpacked.
    velout: optional
        Same as posout, but for the velocities
        
    Returns
    -------
    pos,vel: tuple
        A tuple of the unpacked position and velocity arrays,
        or the number of unpacked particles if an output array
        was given.
    
    '''
    intdata = intdata.reshape(-1)
    assert(intdata.dtype == np.int32)
    N = len(intdata)
    
    if posout is None:
        _posout = np.empty(N, dtype=float_dtype)
    elif posout is False:
        _posout = np.empty(0)
    else:
        _posout = posout.view()
        _posout.shape = -1  # enforces no copy
    
    if velout is None:
        _velout = np.empty(N, dtype=float_dtype)
    elif velout is False:
        _velout = np.empty(0)
    else:
        _velout = velout.view()
        _velout.shape = -1  # enforces no copy
            
    _unpack_rvint(intdata, boxsize, _posout, _velout)
    
    ret = []
    if posout is None:
        ret += [_posout.reshape(N//3,3)]
    elif posout is False:
        ret += [0]
    else:
        ret += [N//3]
        
    if velout is None:
        ret += [_velout.reshape(N//3,3)]
    elif velout is False:
        ret += [0]
    else:
        ret += [N//3]
        
    return tuple(ret)


@nb.njit
def _unpack_rvint(intdata, boxsize, posout, velout):
    '''Helper for unpack_rvint
    '''

    N = len(intdata)
    posscale = boxsize*(2.**-12.)/1e6
    velscale = 6000./2048
    pmask = np.int32(0xfffff000)
    vmask = np.int32(0xfff)
    
    lenp = len(posout)
    lenv = len(velout)
    
    for i in range(N):
        if lenp > 0:
            posout[i] = (intdata[i]&pmask)*posscale
        if lenv > 0:
            velout[i] = ((intdata[i]&vmask) - 2048)*velscale
            

def unpack_pids(packed, box=None, ppd=None, pid=False, lagr_pos=False, tagged=False, density=False, lagr_idx=False, float_dtype=np.float32):
    '''
    Extract fields from bit-packed PIDs.  The PID (really, the 64-bit aux field)
    enocdes the particle ID, the Lagrangian index (and therefore position), the
    density, and the L2 tagged field.
    
    Parameters
    ----------
    packed: array-like of np.uint64, shape (N,)
        The bit-packed PID (i.e. the aux field)
        
    box: float, optional
        The box size, needed only for ``lagr_pos``
        
    ppd: int, optional
        The particles-per-dimension, needed only for ``lagr_pos``
        
    pid: bool, optional
        Whether to unpack and return the unique particle ID.
        
        Loaded as a ``np.int64`` array of shape `(N,)`.
    
    lagr_idx: bool, optional
        Whether to unpack and return the Lagrangian index, which is the `(i,j,k)`
        integer coordinates of the particle in the cubic lattice used in Abacus
        pre-initial conditions. The ``lagr_pos`` field will automatically convert
        this index to a position.
        
        Loaded as a ``np.int16`` array of shape `(N,3)`.
    
    lagr_pos: bool, optional
        Whether to unpack and return the Lagrangian position of the particles,
        based on the Lagrangian index (``lagr_idx``).
        
        Loaded as array of type ``dtype`` and shape `(N,3)`.
    
    tagged: bool, optional
        Whether to unpack and return the CompaSO L2 tagged bit of the particles---
        whether the particle was ever part of an L2 group (i.e. halo core).
        
        Loaded as a ``np.bool8`` array of shape `(N,)`.
    
    density: bool, optional
        Whether to unpack and return the local density estimate, in units of mean
        density.
        
        Loaded as a array of type ``dtype`` and shape `(N,)`.
        
    float_dtype: np.dtype, optional
        The dtype in which to store float arrays. Default: ``np.float32``
        
    Returns
    -------
    unpacked_arrays: dict of ndarray
        A dictionary of all fields that were unpacked
    '''
    packed = np.asanyarray(packed, dtype=np.uint64)
    
    if lagr_pos is not False:
        if box is None:
            raise ValueError('Must supply `box` if requesting `lagr_pos`')
        if ppd is None:
            raise ValueError('Must supply `ppd` if requesting `lagr_pos`')
            
    N = len(packed)
    if not np.isclose(ppd, int(round(ppd))):
        raise ValueError(f'ppd "{ppd}" not valid int?')
    ppd = int(round(ppd))
    
    arr = {}
    if pid is True:
        arr['pid'] = np.empty(N, dtype=np.int64)
    if lagr_pos is True:
        arr['lagr_pos'] = np.empty((N,3), dtype=float_dtype)
    if lagr_idx is True:
        arr['lagr_idx'] = np.empty((N,3), dtype=np.int16)
    if tagged is True:
        arr['tagged'] = np.empty(N, dtype=np.bool8)
    if density is True:
        arr['density'] = np.empty(N, dtype=float_dtype)
        
    _unpack_pids(packed, box, ppd, float_dtype=float_dtype, **arr)
        
    return arr
    

@nb.njit
def _unpack_pids(packed, box, ppd, pid=None, lagr_pos=None, tagged=None, density=None, lagr_idx=None, float_dtype=np.float32):
    '''Helper to extract the Lagrangian position, tagged info, and
    density from the ids of the particles
    '''

    N = len(packed)
    box = np.float64(box)
    inv_ppd = float_dtype(box/ppd)
    half = float_dtype(box/2)

    for i in range(N):
        if lagr_idx is not None:
            lagr_idx[i,0] =  packed[i] & AUXXPID
            lagr_idx[i,1] = (packed[i] & AUXYPID) >> 16
            lagr_idx[i,2] = (packed[i] & AUXZPID) >> 32
            
        if lagr_pos is not None:
            lagr_pos[i,0] = (packed[i] & AUXXPID)*inv_ppd - half
            lagr_pos[i,1] = ((packed[i] & AUXYPID) >> 16)*inv_ppd - half
            lagr_pos[i,2] = ((packed[i] & AUXZPID) >> 32)*inv_ppd - half

        if tagged is not None:
            tagged[i] = (packed[i] >> AUXTAGGED) & 1

        if density is not None:
            density[i] = ((packed[i] & AUXDENS) >> ZERODEN)**2  # max is 2**10, squaring gets to 2**20

        if pid is not None:
            pid[i] = packed[i] & AUXPID



import os

import numpy as np
import matplotlib.pyplot as plt
import argparse
import getdist
from getdist import plots, MCSamples
import yaml

DEFAULTS = {}
DEFAULTS['path2config'] = 'config/abacus_hod.yaml'

def get_samples(outfile, par_names, w_rat, n_par, b_iter):
    marg_chains = np.loadtxt(outfile)
    # uncomment for when your chains have been complete
    #marg_chains = marg_chains[w_rat*n_par*b_iter:]
    marg_chains = marg_chains[3*marg_chains.shape[0]//4:]
    hsc = MCSamples(samples=marg_chains, names=par_names)
    return hsc

def main(path2config):
    # read parameters
    config = yaml.load(open(path2config))
    fit_params = config['fit_params']
    ch_params = config['ch_config_params']

    # parameters
    n_iter = ch_params['sampleIterations']
    w_rat = ch_params['walkersRatio']
    b_iter = ch_params['burninIterations']
    par_names = fit_params.keys()
    lab_names = par_names
    n_par = len(par_names)

    # what are we plotting
    HOD_pars = par_names
    filename = "triangle_test.png"
    dir_chains = ch_params['path2output']

    # walkers ratio, number of params and burn in iterations
    marg_outfile = os.path.join(dir_chains, (ch_params['chainsPrefix']+".txt"))

    # read the samples after removing burnin
    marg_hsc = get_samples(marg_outfile, par_names, w_rat, n_par, b_iter)

    # Triangle plot
    g = plots.getSubplotPlotter()
    g.settings.legend_fontsize = 20
    g.settings.scaling_factor = 0.1
    g.triangle_plot([marg_hsc], params=HOD_pars)
    plt.savefig(filename)
    plt.close()

class ArgParseFormatter(argparse.RawDescriptionHelpFormatter, argparse.ArgumentDefaultsHelpFormatter):
    pass


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description=__doc__, formatter_class=ArgParseFormatter)
    parser.add_argument('--path2config', dest='path2config', type=str, help='Path to config file.', default=DEFAULTS['path2config'])
    
    args = vars(parser.parse_args())    
    main(**args)

